#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Module that stores warnings in dessia_common. """


class SerializationWarning(Warning):
    """ Warning during serialization. """

    def __init__(self, message: str):
        self.message = message

        Warning.__init__(self, message)



"""
Tests for dessia_common.utils.helpers
"""
from dessia_common.models import all_cars_no_feat
from dessia_common.datatools.dataset import Dataset
from dessia_common.utils.helpers import concatenate

values_list = [all_cars_no_feat, all_cars_no_feat]
values_hlist = [Dataset(all_cars_no_feat), Dataset(all_cars_no_feat)]
multi_types = [all_cars_no_feat, Dataset(all_cars_no_feat)]
wrong_type = [1,2,3,4,5,6]

assert(concatenate(values_list) == all_cars_no_feat + all_cars_no_feat)
assert(concatenate(values_hlist) == Dataset(all_cars_no_feat + all_cars_no_feat))

try:
    concatenate(multi_types)
except Exception as e:
    assert(e.args[0] == "Block Concatenate only defined for operands of the same type.")

try:
    concatenate(wrong_type)
except Exception as e:
    assert(e.args[0] == ("Block Concatenate only defined for classes with 'extend' method"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Types tools. """

import warnings
from collections.abc import Iterator, Sequence
from typing import Any, Dict, List, Tuple, Type, Union, get_origin, get_args
import orjson
from dessia_common.abstract import CoreDessiaObject
from dessia_common.typings import InstanceOf, MethodType, ClassMethodType
from dessia_common.files import BinaryFile, StringFile
from dessia_common.schemas.core import TYPING_EQUIVALENCES, union_is_default_value, is_typing, serialize_annotation
from dessia_common.utils.helpers import get_python_class_from_class_name

SIMPLE_TYPES = [int, str]

TYPES_STRINGS = {int: 'int', float: 'float', bool: 'boolean', str: 'str', list: 'list', tuple: 'tuple', dict: 'dict'}

SEQUENCE_TYPINGS = ['List', 'Sequence', 'Iterable']

SERIALIZED_BUILTINS = ['float', 'builtins.float', 'int', 'builtins.int', 'str', 'builtins.str', 'bool', 'builtins.bool']


def is_classname_transform(string: str):
    """ Check if string is class name and return class if yes. """
    if '.' in string:
        split_string = string.split('.')
        if len(split_string) >= 2:
            try:
                class_ = get_python_class_from_class_name(string)
                return class_
            except (AttributeError, TypeError, ModuleNotFoundError, SyntaxError):
                return False
    return False


def is_jsonable(obj):
    """ Returns if the object can be dumped as it is in a JSON. """
    # First trying with orjson which is more efficient
    try:
        orjson.dumps(obj, option=orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_NON_STR_KEYS).decode('utf-8')
        return True
    except Exception:
        return False


def is_serializable(_):
    """ Return True if object is deeply serializable as Dessia's standards, else False. """
    msg = "Function is_serializable has been moved to module serialization.py. Please use this one instead."
    raise NotImplementedError(msg)


def is_sequence(obj) -> bool:
    """
    Return True if object is sequence (but not string), else False.

    :param obj: Object to check
    :return: bool. True if object is a sequence but not a string. False otherwise
    """
    if not hasattr(obj, "__len__") or not hasattr(obj, "__getitem__"):
        # Performance improvements for trivial checks
        return False

    if is_list(obj) or is_tuple(obj):
        # Performance improvements for trivial checks
        return True
    return isinstance(obj, Sequence) and not isinstance(obj, str)


def is_list(obj) -> bool:
    """ Check if given obj is exactly of type list (not instance of). Used mainly for performance. """
    return obj.__class__ == list


def is_tuple(obj) -> bool:
    """ Check if given obj is exactly of type tuple (not instance of). Used mainly for performance. """
    return obj.__class__ == tuple


def is_builtin(type_):
    """ Return True if type_ is a simple python builtin (int, float, bool or str). """
    return type_ in TYPING_EQUIVALENCES


def is_simple(obj):
    """ Return True if given object is a int or a str or None. Used mainly for performance. """
    return obj is None or obj.__class__ in SIMPLE_TYPES


def isinstance_base_types(obj):
    """ Return True if the object is either a str, a float an int or None. """
    if is_simple(obj):
        # Performance improvements for trivial types
        return True
    return isinstance(obj, (str, float, int))


def is_dessia_file(obj):
    """ Wether object inherits from dessia files. """
    return isinstance(obj, (BinaryFile, StringFile))


def is_file_or_file_sequence(object_):
    """ Whether object inherits from dessia file or is a container of such objects. """
    is_file = is_dessia_file(object_)
    if not is_file and is_sequence(object_):
        for obj in object_:
            if is_file_or_file_sequence(obj):
                return True
    return is_file


def unfold_deep_annotation(typing_=None):
    """ Get origin (tuple, list,...) and arguments (type,...) from typing. """
    if is_typing(typing_):
        origin = get_origin(typing_)
        args = get_args(typing_)
        return origin, args
    return None, None


def serialize_typing(typing_):
    """ Compute a string from a type. """
    warnings.warn("Function serialize_typing is deprecated. Use serialize_annotation instead", DeprecationWarning)
    return serialize_annotation(typing_)


def type_from_argname(argname):
    """ Try and compute a type from an argument's name. """
    splitted_argname = argname.rsplit('.', 1)
    if argname:
        if splitted_argname[0] == '__builtins__':
            argname = f"builtins.{splitted_argname[1]}"
        return get_python_class_from_class_name(argname)
    return Any


TYPING_FROM_SERIALIZED_NAME = {"List": List, "Tuple": Tuple, "Iterator": Iterator, "Dict": Dict}


def deserialize_typing(serialized_typing: str):
    """ Compute a typing from a string. """
    # TODO : handling recursive deserialization
    if isinstance(serialized_typing, str):
        # TODO other builtins should be implemented
        if serialized_typing in SERIALIZED_BUILTINS:
            return deserialize_builtin_typing(serialized_typing)

        if serialized_typing in ["dessia_common.files.StringFile", "dessia_common.files.BinaryFile"]:
            return deserialize_file_typing(serialized_typing)

        if serialized_typing in ["dessia_common.typings.MethodType", "dessia_common.typings.ClassMethodType"]:
            return deserialize_method_typing(serialized_typing)

        if serialized_typing == "Type":
            return Type

        if '[' in serialized_typing:
            toptype, remains = serialized_typing.split('[', 1)
            full_argname = remains.rsplit(']', 1)[0]
            if "[" in full_argname:
                return TYPING_FROM_SERIALIZED_NAME[toptype][deserialize_typing(full_argname)]
        else:
            toptype = serialized_typing
            full_argname = ''
        if toptype == 'List':
            return List[type_from_argname(full_argname)]
        if toptype == 'Tuple':
            return deserialize_tuple_typing(full_argname)
        if toptype == 'Dict':
            args = full_argname.split(', ')
            key_type = type_from_argname(args[0])
            value_type = type_from_argname(args[1])
            return Dict[key_type, value_type]
        if toptype == "InstanceOf":
            return InstanceOf[type_from_argname(full_argname)]
        return get_python_class_from_class_name(serialized_typing)
    raise NotImplementedError(f'{serialized_typing} of type {type(serialized_typing)}')


def deserialize_tuple_typing(full_argname):
    """ Compute a tuple typing from a string. """
    if ', ' in full_argname:
        args = full_argname.split(', ')
        if len(args) == 0:
            return Tuple
        if len(args) == 1:
            type_ = type_from_argname(args[0])
            return Tuple[type_]
        if len(set(args)) == 1:
            type_ = type_from_argname(args[0])
            return Tuple[type_, ...]

        raise TypeError("Heterogenous tuples are forbidden as types for workflow non-block variables.")
    return Tuple[type_from_argname(full_argname)]


def deserialize_file_typing(serialized_typing):
    """ Compute a file typing from a string. """
    if serialized_typing == "dessia_common.files.StringFile":
        return StringFile
    if serialized_typing == "dessia_common.files.BinaryFile":
        return BinaryFile
    raise NotImplementedError(f"File typing {serialized_typing} deserialization is not implemented")


def deserialize_method_typing(serialized_typing):
    """ Compute a method typing from a string. """
    if serialized_typing == "dessia_common.typings.MethodType":
        return MethodType
    if serialized_typing == "dessia_common.typings.ClassMethodType":
        return ClassMethodType
    raise NotImplementedError(f"Method typing {serialized_typing} deserialization is not implemented")


def deserialize_builtin_typing(serialized_typing):
    """ Compute a builtin typing from a string. """
    if serialized_typing in ['float', 'builtins.float']:
        return float
    if serialized_typing in ['int', 'builtins.int']:
        return int
    if serialized_typing in ['str', 'builtins.str']:
        return str
    if serialized_typing in ['bool', 'builtins.bool']:
        return bool
    raise NotImplementedError(f"Builtin typing of {serialized_typing} deserialization is not implemented")


def is_bson_valid(value, allow_nonstring_keys=False) -> Tuple[bool, str]:
    """ Return BSON validity (bool) and a hint (str). """
    if isinstance(value, (int, float, str)):
        return True, ''

    if value is None:
        return True, ''

    if isinstance(value, dict):
        for k, subvalue in value.items():
            # Key check
            if isinstance(k, str):
                if '.' in k:
                    return False, f'key {k} of dict is a string containing a ., which is forbidden'
            elif isinstance(k, float):
                return False, f'key {k} of dict is a float, which is forbidden'
            elif isinstance(k, int):
                if not allow_nonstring_keys:
                    log = f'key {k} of dict is an unsuported type {type(k)},' \
                          f'use allow_nonstring_keys=True to allow'
                    return False, log
            else:
                return False, f'key {k} of dict is an unsuported type {type(k)}'

            # Value Check
            v_valid, hint = is_bson_valid(value=subvalue, allow_nonstring_keys=allow_nonstring_keys)
            if not v_valid:
                return False, hint

    elif is_sequence(value):
        for subvalue in value:
            valid, hint = is_bson_valid(value=subvalue, allow_nonstring_keys=allow_nonstring_keys)
            if not valid:
                return valid, hint
    else:
        return False, f'Unrecognized type: {type(value)}'
    return True, ''


# TODO recursive_type and recursive_type functions look weird


def recursive_type(obj):
    """ What is the difference with serialize typing (?). """
    if isinstance(obj, tuple(list(TYPING_EQUIVALENCES.keys()) + [dict])):
        type_ = TYPES_STRINGS[type(obj)]
    elif isinstance(obj, CoreDessiaObject):
        type_ = obj.__module__ + '.' + obj.__class__.__name__
    elif hasattr(obj, 'output_type'):
        type_ = obj.output_type
    elif isinstance(obj, (list, tuple)):
        type_ = []
        for element in obj:
            type_.append(recursive_type(element))
    elif obj is None:
        type_ = None
    else:
        raise NotImplementedError(obj)
    return type_


def typematch(type_: Type, match_against: Type) -> bool:
    """
    Return whether type_ matches against match_against.

    match_against needs to be "wider" than type_, and the check is not bilateral
    """
    # TODO Implement a more intelligent check for Unions : Union[T, U] should match against Union[T, U, V]
    # TODO Implement a check for Dict
    if type_ == match_against or match_against is Any or particular_typematches(type_, match_against):
        # Trivial cases. If types are strictly equal, then it should pass straight away
        return True

    if is_typing(type_):
        return complex_first_type_match(type_, match_against)

    if not is_typing(match_against):
        # type_ and match_against aren't complex : check for subclass only
        if issubclass(type_, match_against):
            return True

    # type_ is not complex and match_against is
    match_against, origin, args = heal_type(match_against)
    if origin is Union:
        matches = [typematch(type_, subtype) for subtype in args]
        return any(matches)
    return False


def complex_first_type_match(type_: Type, match_against: Type) -> bool:
    """ Match type when type_ is a complex typing (List, Union, Tuple,...). """
    # Complex typing for the first type_. Cases : List, Tuple, Union
    if not is_typing(match_against):
        # Type matching is unilateral and match against should be wider than type_
        return False

    # Inspecting and healing types
    type_, type_origin, type_args = heal_type(type_)
    match_against, match_against_origin, match_against_args = heal_type(match_against)

    if type_origin != match_against_origin:
        # Being strict for now. Is there any other case than default values where this would be wrong ?
        return False

    if type_origin is list:
        # Can only have one argument, should match
        return typematch(type_args[0], match_against_args[0])

    if type_origin is tuple:
        # Order matters, all arguments should match
        return all(typematch(a, b) for a, b in zip(type_args, match_against_args))

    if type_origin is dict:
        # key type AND value type should match
        return typematch(type_args[0], match_against_args[0]) and typematch(type_args[1], match_against_args[1])

    if type_origin is Union:
        # type arguments must be a subset of match_against arguments set
        type_argsset = set(type_args)
        match_against_argsset = set(match_against_args)
        return type_argsset.issubset(match_against_argsset)

    # Otherwise, it is not implemented
    raise NotImplementedError(f"Type {type_} is a complex typing and cannot be matched against others yet")


def heal_type(type_: Type):
    """
    Inspect type and returns its parameters.

    For now, only checks whether the type is an 'Optional' / Union[T, NoneType], which should be flattened and not
    considered.

    Returns the cleaned type, origin and arguments.
    """
    type_origin = get_origin(type_)
    type_args = get_args(type_)
    if type_origin is Union:
        # Check for default values false positive
        if union_is_default_value(type_):
            type_ = type_args[0]
            type_origin = get_origin(type_)
            type_args = get_args(type_)
    return type_, type_origin, type_args


def particular_typematches(type_: Type, match_against: Type) -> bool:
    """ Check for specific cases of type matches and return a Boolean. """
    if type_ is int and match_against is float:
        return True
    # Not refactoring this as a one-liner for now, as more cases should be added in the future.
    return False



""" Provides decorators that work as 'flags' for display settings. """

from typing import Type, List, TypeVar
import inspect
import ast
import textwrap


DISPLAY_DECORATORS = ["plot_data_view", "markdown_view", "cad_view"]


def get_all_decorated_methods(class_: Type) -> List[ast.FunctionDef]:
    """ Get all decorated method from class_. """
    methods = inspect.getmembers(class_, inspect.isfunction) + inspect.getmembers(class_, inspect.ismethod)
    function_defs = []
    for _, method in methods:
        source = textwrap.dedent(inspect.getsource(method))
        method_tree = ast.parse(source)
        function_defs.append(method_tree.body[0])
    return [m for m in function_defs if isinstance(m, ast.FunctionDef) and m.decorator_list]


def get_decorated_methods(class_: Type, decorator_name: str):
    """ Get all method from class_ that are decorated with decorator_name. """
    methods = get_all_decorated_methods(class_)
    method_names = []

    for method in methods:
        decorator_names = []
        for decorator in method.decorator_list:
            if isinstance(decorator, ast.Call):
                name = decorator.func.attr if isinstance(decorator.func, ast.Attribute) else decorator.func.id
            else:
                name = decorator.attr if isinstance(decorator, ast.Attribute) else decorator.id
            decorator_names.append(name)

        if decorator_name in decorator_names:
            method_names.append(method.name)
    return [getattr(class_, n) for n in method_names]


def plot_data_view(selector: str = None, load_by_default: bool = False):
    """
    Decorator to plot data.

    :param str selector: A custom and unique name that identifies the display.
        It is what is displayed on platform to select your view.

    :param bool load_by_default: Whether the view should be displayed on platform by default or not.
    """
    def decorator(function):
        """ Decorator to plot data."""
        set_decorated_function_metadata(function=function, type_="plot_data", selector=selector,
                                        serialize_data=True, load_by_default=load_by_default)
        return function
    return decorator


def markdown_view(selector: str, load_by_default: bool = False):
    """
    Decorator to markdown.

    :param str selector: Unique name that identifies the display.
        It is what is displayed on platform to select your view.

    :param bool load_by_default: Whether the view should be displayed on platform by default or not.
    """
    def decorator(function):
        """ Decorator to markdown. """
        set_decorated_function_metadata(function=function, type_="markdown", selector=selector,
                                        serialize_data=False, load_by_default=load_by_default)
        return function
    return decorator


def cad_view(selector: str, load_by_default: bool = False):
    """
    Decorator to 3D views.

    :param str selector: Unique name that identifies the display.
        It is what is displayed on platform to select your view.

    :param bool load_by_default: Whether the view should be displayed on platform by default or not.
    """
    def decorator(function):
        """ Decorator to 3D views. """
        set_decorated_function_metadata(function=function, type_="babylon_data", selector=selector,
                                        serialize_data=True, load_by_default=load_by_default)
        return function
    return decorator


def set_decorated_function_metadata(function, type_: str, selector: str = None,
                                    serialize_data: bool = False, load_by_default: bool = False):
    """ Attach metadata to function object. Is there any better way to do this ? It seems a bit dirty. """
    setattr(function, "type_", type_)
    setattr(function, "selector", selector)
    setattr(function, "serialize_data", serialize_data)
    setattr(function, "load_by_default", load_by_default)


T = TypeVar("T")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Nov 17 08:07:43 2021

@author: steven
"""

import dessia_common.utils.algebra as dc_alg

print(dc_alg.number2factor(12))
print(dc_alg.number3factor(12))

print(dc_alg.get_incomplete_factors(12, dc_alg.number2factor(6)[0]))



from dessia_common.workflow.core import Workflow, Variable, Pipe, NAME_VARIABLE, RESULT_VARIABLE_NAME
from dessia_common.workflow.blocks import InstantiateModel
from dessia_common.forms import Generator

# Empty Workflow with only the init computed name variable
empty_workflow = Workflow(blocks=[], pipes=[], output=None)
empty_workflow._check_platform()
assert len(empty_workflow.detached_variables) == 1
assert empty_workflow.detached_variables[0].name == RESULT_VARIABLE_NAME
assert len(empty_workflow.nonblock_variables) == 0

empty_dict = empty_workflow.to_dict()
assert len(empty_dict["nonblock_variables"]) == 1
assert empty_dict["nonblock_variables"][0]["name"] == RESULT_VARIABLE_NAME

# Workflow with one block and the detached init computed name variable
block = InstantiateModel(model_class=Generator, name="Generator")
lone_block_workflow = Workflow(blocks=[block], pipes=[], output=None)
lone_block_workflow._check_platform()
assert len(lone_block_workflow.detached_variables) == 1
assert lone_block_workflow.detached_variables[0].name == RESULT_VARIABLE_NAME
assert len(lone_block_workflow.nonblock_variables) == 0
assert len(lone_block_workflow.inputs) == 4  # TODO Should the detached variable be seen as an input ?

lone_block_dict = lone_block_workflow.to_dict()
assert len(empty_dict["nonblock_variables"]) == 1  # Detached variables are serialized as nonblock
assert empty_dict["nonblock_variables"][0]["name"] == RESULT_VARIABLE_NAME

# Workflow with wired block and name variable. Name var should be nonblock and not detached anymore
pipe_name = Pipe(input_variable=NAME_VARIABLE, output_variable=lone_block_workflow.inputs[3])
wired_workflow = Workflow(blocks=[block], pipes=[pipe_name], output=None)
wired_workflow._check_platform()
assert len(wired_workflow.detached_variables) == 0
assert len(wired_workflow.nonblock_variables) == 1
assert wired_workflow.nonblock_variables[0].name == RESULT_VARIABLE_NAME
assert len(wired_workflow.inputs) == 4

wired_dict = wired_workflow.to_dict()
assert len(wired_dict["nonblock_variables"]) == 1
assert wired_dict["nonblock_variables"][0]["name"] == RESULT_VARIABLE_NAME

# Workflow with an user variable added as a detached one
user_variable = Variable(type_=int, default_value="2", name="User Parameter")
user_variable_workflow = Workflow(blocks=[block], pipes=[pipe_name], output=None, detached_variables=[user_variable])
user_variable_workflow._check_platform()
assert len(user_variable_workflow.detached_variables) == 1
assert user_variable_workflow.detached_variables[0].name == "User Parameter"
assert len(user_variable_workflow.nonblock_variables) == 1
assert user_variable_workflow.nonblock_variables[0].name == RESULT_VARIABLE_NAME
assert len(user_variable_workflow.inputs) == 4

user_variable_dict = user_variable_workflow.to_dict()
assert len(user_variable_dict["nonblock_variables"]) == 2
assert user_variable_dict["nonblock_variables"][0]["name"] == RESULT_VARIABLE_NAME
assert user_variable_dict["nonblock_variables"][1]["name"] == "User Parameter"

# Unwired Workflow with an user variable added as a detached one
double_variable_workflow = Workflow(blocks=[block], pipes=[], output=None, detached_variables=[user_variable])
double_variable_workflow._check_platform()
assert len(double_variable_workflow.detached_variables) == 2
assert double_variable_workflow.detached_variables[0].name == RESULT_VARIABLE_NAME
assert double_variable_workflow.detached_variables[1].name == "User Parameter"
assert len(double_variable_workflow.nonblock_variables) == 0
assert len(double_variable_workflow.inputs) == 4

double_variable_dict = double_variable_workflow.to_dict()
assert len(double_variable_dict["nonblock_variables"]) == 2
assert double_variable_dict["nonblock_variables"][0]["name"] == RESULT_VARIABLE_NAME
assert double_variable_dict["nonblock_variables"][1]["name"] == "User Parameter"

# Workflow with name variable that isn't connected and user variable is
pipe_var = Pipe(input_variable=user_variable, output_variable=block.inputs[0])
switched_user_variable_workflow = Workflow(blocks=[block], pipes=[pipe_var], output=None,
                                           detached_variables=[user_variable])
switched_user_variable_workflow._check_platform()
assert len(switched_user_variable_workflow.detached_variables) == 1
assert switched_user_variable_workflow.detached_variables[0].name == RESULT_VARIABLE_NAME
assert len(switched_user_variable_workflow.nonblock_variables) == 1
assert switched_user_variable_workflow.nonblock_variables[0].name == "User Parameter"
assert len(switched_user_variable_workflow.inputs) == 4

switched_user_variable_dict = switched_user_variable_workflow.to_dict()
assert len(switched_user_variable_dict["nonblock_variables"]) == 2
assert switched_user_variable_dict["nonblock_variables"][0]["name"] == "User Parameter"
assert switched_user_variable_dict["nonblock_variables"][1]["name"] == RESULT_VARIABLE_NAME

# Variante of last check with user variable not defined as detached variable
switched_user_variable_workflow = Workflow(blocks=[block], pipes=[pipe_var], output=None)
switched_user_variable_workflow._check_platform()
assert len(switched_user_variable_workflow.detached_variables) == 1
assert switched_user_variable_workflow.detached_variables[0].name == RESULT_VARIABLE_NAME
assert len(switched_user_variable_workflow.nonblock_variables) == 1
assert switched_user_variable_workflow.nonblock_variables[0].name == "User Parameter"
assert len(switched_user_variable_workflow.inputs) == 4

switched_user_variable_dict = switched_user_variable_workflow.to_dict()
assert len(switched_user_variable_dict["nonblock_variables"]) == 2
assert switched_user_variable_dict["nonblock_variables"][0]["name"] == "User Parameter"
assert switched_user_variable_dict["nonblock_variables"][1]["name"] == RESULT_VARIABLE_NAME

# Complete Workflow
another_user_variable = Variable(type_=int, default_value=5, name="Number of Solutions")
pipe_numb = Pipe(input_variable=another_user_variable, output_variable=block.inputs[1])
complete_workflow = Workflow(blocks=[block], pipes=[pipe_name, pipe_var, pipe_numb],
                             output=block.outputs[0], detached_variables=[user_variable])
complete_workflow._check_platform()
assert len(complete_workflow.detached_variables) == 0
assert len(complete_workflow.nonblock_variables) == 3
assert complete_workflow.nonblock_variables[0].name == RESULT_VARIABLE_NAME
assert complete_workflow.nonblock_variables[1].name == "User Parameter"
assert complete_workflow.nonblock_variables[2].name == "Number of Solutions"
assert len(complete_workflow.inputs) == 4

complete_dict = complete_workflow.to_dict()
assert len(complete_dict["nonblock_variables"]) == 3
assert complete_dict["nonblock_variables"][0]["name"] == RESULT_VARIABLE_NAME
assert complete_dict["nonblock_variables"][1]["name"] == "User Parameter"
assert complete_dict["nonblock_variables"][2]["name"] == "Number of Solutions"


print("workflow_building.py script has passed")



from dessia_common.tests import System
from dessia_common.workflow.blocks import InstantiateModel, ModelMethod
from dessia_common.typings import MethodType
from dessia_common.workflow.core import Pipe, Workflow

documentation = """"""

block_0 = InstantiateModel(model_class=System, name="Instantiate Generator", position=(0, 0))
block_1 = ModelMethod(method_type=MethodType(System, 'power_simulation'), name="Generator Generate", position=(0, 0))
blocks = [block_0, block_1]


pipe_0 = Pipe(block_0.outputs[0], block_1.inputs[0])
pipes = [pipe_0]

workflow = Workflow(blocks, pipes, output=block_1.outputs[0], documentation=documentation, name="")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Schema generation functions. """
import re
import warnings
from copy import deepcopy
import inspect
import collections.abc
from typing import Tuple, Dict, List, Type, get_args, get_origin, get_type_hints, Callable, Union, \
    TypeVar, TypedDict, Optional, Any
from functools import cached_property
from dessia_common.utils.helpers import full_classname, get_python_class_from_class_name
from dessia_common.abstract import CoreDessiaObject
from dessia_common.displays import DisplayObject
from dessia_common.files import BinaryFile, StringFile
from dessia_common.typings import (MethodType, ClassMethodType, InstanceOf, Subclass, AttributeType, ClassAttributeType,
                                   CadViewType, PlotDataType, MarkdownType, ViewType)
from dessia_common.measures import Measure
from dessia_common.utils.helpers import prettyname
from dessia_common.schemas.interfaces import Annotations, T, PropertySchema
from dessia_common.checks import CheckList, FailedCheck, PassedCheck, CheckWarning

SCHEMA_HEADER = {"definitions": {}, "$schema": "http://json-schema.org/draft-07/schema#",
                 "type": "object", "required": [], "properties": {}}
RESERVED_ARGNAMES = ["self", "cls", "progress_callback", "return"]
TYPING_EQUIVALENCES = {int: "number", float: "number", bool: "boolean", str: "string"}
TYPES_FROM_STRING = {"unicode": str, "str": str, "float": float, "int": int, "bool": bool}

_fullargsspec_cache = {}

UNDEFINED = object()

EMPTY_PARSED_ATTRIBUTE = {"desc": "", "type": "", "annotation": ""}
FAILED_DOCSTRING_PARSING = {"description": "Docstring parsing failed", "attributes": {}}
FAILED_ATTRIBUTE_PARSING = {"desc": "Attribute documentation parsing failed", "type": "", "annotation": ""}


class BackendReference(CoreDessiaObject):
    """ Enable Backend reference injection. """

    def __init__(self, object_id: str):
        self.object_id = object_id

    def to_dict(self, use_pointers: bool = True, memo=None, path: str = "#", id_method: bool = True, id_memo=None):
        """ Is used for default value when computing schema to dict with objects default values. """
        return {"object_id": self.object_id}


class UntypedArgument(FailedCheck):
    """ Used when an argument is not typed in function or class definition. """

    def __init__(self, attribute: str):
        super().__init__(f"Attribute '{attribute}' : has no typing")


class WrongNumberOfArguments(FailedCheck):
    """ Used when a typing does not have the right amount of arguments. """


class WrongType(FailedCheck):
    """ Used when an annotation does not have the right type. """


class UnsupportedDefault(FailedCheck):
    """ Used when an argument defines a default that is not supported. """


class ParsedAttribute(TypedDict):
    """ Parsed description of a docstring attribute. """

    desc: str
    type_: str
    annotation: str


class ParsedDocstring(TypedDict):
    """ Parsed description of a docstring. """

    description: str
    attributes: Dict[str, ParsedAttribute]


class SchemaAttribute:
    """
    A wrapper that connects a schema attribute name with its attached data.

    :param name: Name of the attribute as defined in the parent schema property
    :param default_value: UNDEFINED if no default value is set in user code.
        None can be a valid default value and should be handled as such.
    :param title: A pretty label for the attribute that is to be shown as input title in frontend.
    :param editable: Whether attribute should be editable in frontend or not. Can be dependent on context.
    :param documentation: Parsed documentation for given attribute.
    """

    def __init__(self, name: str, default_value: T = UNDEFINED, title: str = "",
                 editable: bool = True, documentation: ParsedAttribute = None):
        self.name = name
        self.default_value = default_value
        self._title = title
        self.editable = editable
        if documentation is None:
            documentation = EMPTY_PARSED_ATTRIBUTE
        self.documentation = documentation

    def __str__(self):
        return self.name

    @cached_property
    def has_default_value(self) -> bool:
        """ Helper property that indicates if default value should be trusted as such or is undefined. """
        return self.default_value is not UNDEFINED

    @cached_property
    def title(self) -> str:
        """ Generate a pretty label from name to not let it be empty. """
        if self._title == "":
            return prettyname(self.name)
        return self._title

    @cached_property
    def description(self) -> str:
        """ Helper to get documentation description. """
        return self.documentation["desc"]

    def to_dict(self):
        """ Base attribute dict. """
        return {"title": self.title, "editable": self.editable, "description": self.description}


class Schema:
    """
    Abstraction of a Schema.

    It takes annotations written as a dict and then writes into a Dict the recursive structure of an object
    that can be handled by dessia_common.
    This dictionary can then be translated as a json to be read by the frontend in order to compute edit forms,
    for example.

    Right now Schema doesn't inherit from any DessiaObject class (SerializableObject ?), but could, in the future.
    That is why it implements methods with the same name.
    """

    def __init__(self, annotations: Annotations, attributes: List[SchemaAttribute],
                 documentation: str = "", name: str = ""):
        self.annotations = annotations
        self.attributes = attributes
        self.documentation = documentation
        self.name = name

        self.required = [a for a in attributes if not a.has_default_value]

        self.property_schemas = {}
        for attribute in self.attributes:
            if attribute.name in self.annotations:
                annotation = self.annotations[attribute.name]
                schema = get_schema(annotation=annotation, attribute=attribute)
                self.property_schemas[attribute.name] = schema

    @property
    def editable_attributes(self) -> List[str]:
        """ Attributes that are not in RESERVED_ARGNAMES. """
        return [a.name for a in self.attributes if a.name not in RESERVED_ARGNAMES]

    def chunk(self, attribute: SchemaAttribute) -> Dict[str, Any]:
        """ Extract and compute a schema from one of the attributes. """
        schema = self.property_schemas.get(attribute.name, None)

        if schema is not None:
            return schema.to_dict()

        # if attribute in self.default_arguments:
        #     # TODO Could use this and Optional proxy in order to inject real default values for mutable
        #     default = self.default_arguments.get(attribute, None)
        #     print("Default", default)
        #     chunk["default_value"] = schema.default_value(definition_default=default)
        return {}

    @property
    def chunks(self) -> List[Dict[str, Any]]:
        """ Concatenate schema chunks into a List. """
        return [self.chunk(a) for a in self.attributes]

    @property
    def standalone_properties(self) -> List[str]:
        """ Return all properties that are standalone. """
        return [a.name for a in self.attributes if self.property_schemas[a.name].standalone_in_db]

    def to_dict(self, **kwargs) -> Dict[str, Any]:
        """ Base Schema. kwargs are added to result as well. """
        schema = deepcopy(SCHEMA_HEADER)
        properties = {a.name: self.chunk(a) for a in self.attributes}
        required = [a.name for a in self.required]
        schema.update({"required": required, "properties": properties, "description": self.documentation})
        schema.update(**kwargs)
        return schema

    def default_dict(self) -> Dict[str, Any]:
        """
        Compute global default Dict.

        If a definition default have been set by user, most schemas will return this value (or serialized).
        if not, schemas will compute a default compatible with platform (None most of the time).
        """
        return {a: s.default_value() for a, s in self.property_schemas.items()}

    def default_value(self) -> Dict[str, Any]:
        """
        A user computed default value.

        It differs from 'default_dict', as only user defined default_value are set here.
        In 'default_dict' method, every property is set.
        """
        return {a: s.default_value() for a, s in self.property_schemas.items() if s.has_default_value}

    def check_list(self) -> CheckList:
        """
        Browse all properties and List potential issues.

        Checks performed for each argument :
        - Is typed in method definition
        - Schema specific check
        """
        issues = CheckList([])

        for attribute in self.attributes:
            # Is typed
            is_typed_check = self.attribute_is_annotated(attribute.name)
            issues += CheckList([is_typed_check])

            if is_typed_check.level != "error":
                # Specific check
                schema = self.property_schemas[attribute.name]
                issues += schema.check_list()
        return issues

    @property
    def is_valid(self) -> bool:
        """ Return whether the class definition is valid or not. """
        return not self.check_list().checks_above_level("error")

    def attribute_is_annotated(self, attribute_name: str) -> PassedCheck:
        """ Check whether given attribute is annotated in function definition or not. """
        if attribute_name not in self.annotations:
            return UntypedArgument(attribute_name)
        return PassedCheck(f"Attribute '{attribute_name}' : is annotated")

    @classmethod
    def from_type(cls, annotation: Type[T], attribute_name: str) -> 'Schema':
        """
        Compute a full schema (not property) from an annotation.

        It is useful in order to provide forms for object of a single type.
        """
        annotations = {attribute_name: annotation}
        attribute = SchemaAttribute(attribute_name)
        return cls(annotations=annotations, attributes=[attribute])

    @classmethod
    def from_serialized_type(cls, type_: str, attribute_name: str) -> 'Schema':
        """ Compute a Schema directly from the serialized value of the annotation. """
        annotation = deserialize_annotation(type_)
        return cls.from_type(annotation=annotation, attribute_name=attribute_name)


class MemberSchema(Schema):
    """
    Abstraction of a Schema computed from a member (Object or Method).

    It reads the user-defined type hints and then writes into a Dict the recursive structure of an object
    that can be handled by dessia_common.
    This dictionary can then be translated as a json to be read by the frontend in order to compute frontend forms,
    for example.

    Right now Schema doesn't inherit from any DessiaObject class (SerializableObject ?), but could, in the future.
    That is why it implements methods with the same name.
    """

    def __init__(self, annotations: Annotations, argspec: inspect.FullArgSpec, docstring: str, name: str = ""):
        attribute_names = [a for a in argspec.args if a not in RESERVED_ARGNAMES]

        try:  # Parse docstring
            parsed_docstring = parse_docstring(docstring=docstring, annotations=annotations)
        except Exception:  # Catching broad exception because we don't want it to break platform if a failure occurs
            parsed_docstring = FAILED_DOCSTRING_PARSING
        documentation = parsed_docstring["description"]
        descriptions = parsed_docstring["attributes"]

        _, default_arguments = split_default_args(argspecs=argspec, merge=False)

        attributes = []
        required = []
        for attribute_name in attribute_names:
            attribute_documentation = descriptions.get(attribute_name, EMPTY_PARSED_ATTRIBUTE)
            attribute = SchemaAttribute(name=attribute_name, documentation=attribute_documentation)
            if attribute_name in default_arguments:
                attribute.default_value = default_arguments[attribute_name]
            else:
                required.append(attribute)
            attributes.append(attribute)

        super().__init__(annotations=annotations, attributes=attributes, documentation=documentation, name=name)

        self.required = required


class ClassSchema(MemberSchema):
    """
    Schema of a class.

    Class must be a subclass of DessiaObject. It reads the __init__ annotations.
    """

    def __init__(self, class_: Type[CoreDessiaObject]):
        self.class_ = class_
        self.python_typing = full_classname(class_, compute_for="class")
        annotations = get_type_hints(class_.__init__)

        members = inspect.getfullargspec(self.class_.__init__)
        docstring = class_.__doc__
        classname = full_classname(class_)

        super().__init__(annotations=annotations, argspec=members, docstring=docstring, name=classname)

    @property
    def editable_attributes(self) -> List[str]:
        """ Attributes that are not in RESERVED_ARGNAMES nor defined as non-editable by user. """
        attributes = super().editable_attributes
        return [a for a in attributes if a not in self.class_._non_editable_attributes]

    @property
    def standalone_in_db(self) -> bool:
        """ Return True if class is standalone. """
        return self.class_._standalone_in_db

    def to_dict(self, **kwargs) -> Dict[str, Any]:
        """ Write the whole schema. """
        schema = super().to_dict(**kwargs)
        classname = full_classname(object_=self.class_, compute_for="class")
        schema.update({"classes": [classname], "standalone_in_db": self.standalone_in_db})
        # TODO Check if we can get rid of this and use Property schema to simplify frontend
        return schema

    def default_dict(self) -> Dict[str, Any]:
        """ Compute class default Dict. Add object_class to base one. """
        dict_ = super().default_dict()
        dict_["object_class"] = self.python_typing
        return dict_


class MethodSchema(MemberSchema):
    """
    Schema of a method.

    Given method should be one of a DessiaObject. It reads its annotations.
    """

    def __init__(self, method: Callable):
        if isinstance(method, property):
            method = method.fget
        self.method = method

        annotations = get_type_hints(method)
        members = inspect.getfullargspec(method)
        self.return_annotation = annotations.get("return", None)
        docstring = method.__doc__
        super().__init__(annotations=annotations, argspec=members, docstring=docstring, name=method.__name__)

        self.required_indices = [str(self.attributes.index(a)) for a in self.required]

    @property
    def serialized(self) -> Dict[str, str]:
        """ Get a dictionary with serialized annotations of method arguments. """
        return {k: s.serialized for k, s in self.property_schemas.items()}

    @property
    def return_schema(self) -> 'Property':
        """ Get the schema of the return annotation. """
        attribute = SchemaAttribute(name="return", editable=False)
        return get_schema(annotation=self.return_annotation, attribute=attribute)

    @property
    def return_serialized(self):
        """ Get the serialized annotation of the return annotation. """
        try:
            return self.return_schema.serialized
        except NotImplementedError:
            return None

    def to_dict(self, **kwargs):
        """ Write the whole schema. """
        schema = deepcopy(SCHEMA_HEADER)
        properties = {str(i): self.chunk(a) for i, a in enumerate(self.attributes)}
        schema.update({"required": self.required_indices, "properties": properties, "description": self.documentation})
        return schema

    def definition_json(self):
        """ Dictionary that denotes the good definition of the method. Useful for low-code features. """
        return {"name": self.name, "return": self.return_serialized, "arguments": self.serialized,
                "valid": self.is_valid, "checks": self.check_list().to_dict()["checks"]}

    def check_list(self) -> CheckList:
        """
        Browse all properties and List potential issues.

        Checks performed for each argument :
        - Is typed in method definition
        - Schema specific check
        """
        issues = super().check_list()
        issues += CheckList([self.return_is_annotated(), self.return_type_is_valid()])
        return issues

    def return_is_annotated(self) -> PassedCheck:
        """ Check whether method return is annotated in definition or not. """
        if "return" not in self.annotations:
            return CheckWarning("Method return : is not annotated")
        return PassedCheck("Method return : is annotated")

    def return_type_is_valid(self) -> PassedCheck:
        """ Check whether given attribute is annotated in function definition or not. """
        try:
            _ = self.return_schema
            return PassedCheck(f"Method return : '{self.return_annotation}' is valid")
        except NotImplementedError:
            return CheckWarning(f"Method return : '{self.return_annotation} is not valid")


class Property:
    """ Base class for a schema property. """

    def __init__(self, annotation: Type[T], attribute: SchemaAttribute):
        self.annotation = annotation
        self.attribute = attribute

    @property
    def schema(self):
        """ Return a reference to itself. Might be overwritten for proxy such as Optional or Annotated. """
        return self

    @cached_property
    def is_file_related(self) -> bool:
        """ Wether the property is related to files. Base property never is. """
        return False

    @cached_property
    def serialized(self) -> str:
        """ Stringified annotation. """
        if isinstance(self.annotation, type):
            # TODO Hot fix for DisplayObjects (and other non dessia_object annotations). Is this OK ?
            return full_classname(self.annotation, compute_for="class")
        return str(self.annotation)

    @cached_property
    def pretty_annotation(self) -> str:
        """ Stringified pretty annotation. """
        return self.serialized

    @property
    def standalone_in_db(self) -> bool:
        """ Properties cannot be in database by default. """
        return False

    @property
    def has_default_value(self):
        """ Helper property that indicates if default value should be trusted as such or is undefined. """
        return self.attribute.default_value != UNDEFINED

    @classmethod
    def annotation_from_serialized(cls, serialized: str):
        """ Simple get class from its name. """
        return get_python_class_from_class_name(serialized)

    @property
    def check_prefix(self) -> str:
        """ Shortcut for Check message prefixes. """
        return f"Attribute '{self.attribute}' : "

    def to_dict(self) -> PropertySchema:
        """ Write base schema as a Dict. """
        attribute_dict = self.attribute.to_dict()
        dict_ = {"python_typing": self.serialized, "type": None, **attribute_dict}
        if self.has_default_value:
            dict_["default_value"] = self.default_value()
        return dict_

    def default_value(self):
        """ Generic default. Yield user default if defined. """
        return self.attribute.default_value

    def inject_reference(self, object_id: str):
        """ Exposed to backend in order to inject MongoDB reference in place of object default value. """
        if self.standalone_in_db:
            self.attribute.default_value = BackendReference(object_id)

    def check_list(self) -> CheckList:
        """
        Check validity of Property Type Hint.

        Checks performed : None. TODO ?
        """
        return CheckList([])


class TypingProperty(Property):
    """ Schema class for typing based annotations. """

    SERIALIZED_REGEXP = r"([^\[\]]*)\[(.*)\]"

    def __init__(self, annotation: Type[T], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

    @property
    def args(self) -> Tuple[Type[T], ...]:
        """ Return Typing arguments. """
        return get_args(self.annotation)

    @property
    def origin(self) -> Type:
        """ Return Typing origin. """
        return get_origin(self.annotation)

    @property
    def args_schemas(self) -> List[Property]:
        """ Get schema for each argument. """
        schemas = []
        for i, arg in enumerate(self.args):
            subattribute = SchemaAttribute(name=f"{self.attribute.name}/{i}", title=f"{self.attribute.title}/{i}",
                                           editable=self.attribute.editable)
            schemas.append(get_schema(annotation=arg, attribute=subattribute))
        return schemas

    @cached_property
    def is_file_related(self) -> bool:
        """ Wether the property is related to files. Is related to files if any of its argument is. """
        return any(s.is_file_related for s in self.args_schemas)

    @cached_property
    def serialized(self) -> str:
        """ Recursively stringify annotation. """
        serialized = self.origin.__name__
        if serialized in ["list", "dict", "tuple", "type"]:
            # TODO Dirty quick-fix. Find a generic way to automatize this
            serialized = serialized.capitalize()
        if self.args:
            return compute_typing_schema_serialization(serialized_typing=serialized, args_schemas=self.args_schemas)
        return serialized

    @cached_property
    def pretty_annotation(self) -> str:
        """ Recursively prettify annotation. """
        prettyfied = self.origin.__name__
        if prettyfied in ["list", "dict", "tuple", "type"]:
            # TODO Dirty quick-fix. Find a generic way to automatize this
            prettyfied = prettyfied.capitalize()
        if self.args:
            return compute_pretty_schema_annotation(serialized_typing=prettyfied, args_schemas=self.args_schemas)
        return prettyfied

    @classmethod
    def annotation_from_serialized(cls, serialized: str) -> Type[T]:
        """ Split Typing and Arguments and delegate deserialization to specific classes. """
        typename = cls.type_from_serialized(serialized)
        schema_class = SERIALIZED_TO_SCHEMA_CLASS[typename]
        return schema_class.annotation_from_serialized(serialized)

    @classmethod
    def type_from_serialized(cls, serialized: str) -> str:
        """ Get Typing from serialized value. """
        return re.match(cls.SERIALIZED_REGEXP, serialized).group(1)

    @classmethod
    def _raw_args_from_serialized(cls, serialized: str) -> str:
        """ Get arguments as str from serialized value. """
        if "[" in serialized and "]" in serialized:
            args = re.match(cls.SERIALIZED_REGEXP, serialized).group(2)
            return args.replace(" ", "")
        return ""

    @classmethod
    def _args_from_serialized(cls, serialized: str) -> Tuple[Type[T], ...]:
        """ Deserialize arguments. """
        rawargs = cls._raw_args_from_serialized(serialized)
        args = extract_args(rawargs)
        return tuple(deserialize_annotation(a) for a in args)

    @classmethod
    def unfold_serialized_annotation(cls, serialized: str):
        """ Get Typing and Arguments as strings. """
        return re.match(cls.SERIALIZED_REGEXP, serialized).groups()

    def has_one_arg(self) -> PassedCheck:
        """ Annotation should have exactly one argument. """
        if len(self.args) != 1:
            pretty_origin = prettyname(self.origin.__name__)
            msg = f"{self.check_prefix}is typed as a '{pretty_origin}' which requires exactly 1 argument. " \
                  f"Expected '{pretty_origin}[T]', got '{self.annotation}'."
            return WrongNumberOfArguments(msg)
        return PassedCheck(f"{self.check_prefix}has exactly one argument in its definition.")


class ProxyProperty(TypingProperty):
    """
    Schema Class for Proxies.

    Proxies are just intermediary types with actual schemas in its arguments. For example OptionalProperty proxy.
    """

    def __init__(self, annotation: Type[T], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

        self.annotation = self.args[0]

    @property
    def schema(self) -> Property:
        """ Return a reference to its only argument. """
        return get_schema(annotation=self.annotation, attribute=self.attribute)

    def to_dict(self) -> Dict[str, Any]:
        """ Write Proxy as a Dict. """
        return self.schema.to_dict()

    @cached_property
    def serialized(self) -> str:
        """ Stringify under-proxy-annotation. """
        return self.schema.serialized

    @cached_property
    def pretty_annotation(self) -> str:
        """ Stringify under-proxy-annotation. """
        return self.schema.pretty_annotation

    @classmethod
    def annotation_from_serialized(cls, serialized: str) -> Type[T]:
        """ Deserialize Proxy annotation. """
        raise NotImplementedError(f"Cannot deserialize annotation '{serialized}' as Proxy.")


class OptionalProperty(ProxyProperty):
    """
    Proxy Schema class for OptionalProperty properties.

    OptionalProperty is only a catch for arguments that default to None.
    Arguments with default values other than None are not considered Optional.
    """

    def __init__(self, annotation: Type[T], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

    @classmethod
    def annotation_from_serialized(cls, serialized: str) -> Type[T]:
        """ Deserialize Optional annotation. """
        raise NotImplementedError("Optional property deserialization not implemented yet. ")


class AnnotatedProperty(ProxyProperty):
    """
    Proxy Schema class for annotated type hints.

    AnnotatedProperty annotations are type hints with more arguments passed, such as value ranges,
    or probably enumerations, precision,...

    This could enable quite effective type checking on frontend form.

    Only available with python >= 3.11
    """

    _not_implemented_msg = "AnnotatedProperty type hints are not implemented yet. This needs python 3.11 at least. " \
                           "Dessia only supports python 3.9 at the moment."

    # TODO Whenever Dessia decides to upgrade to python 3.11
    def __init__(self, annotation: Type[T], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)
        raise NotImplementedError(self._not_implemented_msg)

    @classmethod
    def annotation_from_serialized(cls, serialized: str) -> Type[T]:
        """ Deserialize Annotated annotation. """
        raise NotImplementedError(cls._not_implemented_msg)

    def to_dict(self) -> Dict[str, Any]:
        """ Write Annotated as a Dict. """
        raise NotImplementedError(self._not_implemented_msg)

    def check_list(self) -> CheckList:
        """
        Check validity of DynamicDict Type Hint.

        Checks performed : None. TODO : Argument validity
        """
        raise NotImplementedError(self._not_implemented_msg)


Builtin = Union[str, bool, float, int]


class BuiltinProperty(Property):
    """ Schema class for Builtin type hints. """

    def __init__(self, annotation: Type[Builtin], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

    @cached_property
    def serialized(self) -> str:
        """ Builtin name. """
        return self.annotation.__name__

    @classmethod
    def annotation_from_serialized(cls, serialized: str) -> Type[T]:
        """ Get real Type from types dictionary. """
        return TYPES_FROM_STRING[serialized]

    def to_dict(self) -> Dict[str, Any]:
        """ Write Builtin as a Dict. """
        chunk = super().to_dict()
        chunk["type"] = TYPING_EQUIVALENCES[self.annotation]
        if self.has_default_value:
            chunk["default_value"] = self.default_value()
        return chunk

    def default_value(self):
        """ Force None default value when undefined. """
        if self.has_default_value:
            return self.attribute.default_value
        return None


class MeasureProperty(BuiltinProperty):
    """ Schema class for Measure type hints. """

    def __init__(self, annotation: Type[Measure], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

    @cached_property
    def serialized(self) -> str:
        """ Full class name. """
        return full_classname(object_=self.annotation, compute_for="class")

    @cached_property
    def pretty_annotation(self) -> str:
        """ Full class name. """
        return self.annotation.__name__

    @classmethod
    def annotation_from_serialized(cls, serialized: str) -> Type[T]:
        """ Deserialize as annotation custom class. """
        return Property.annotation_from_serialized(serialized)

    def to_dict(self) -> Dict[str, Any]:
        """ Write Measure as a Dict. """
        chunk = Property.to_dict(self)
        chunk.update({"si_unit": self.annotation.si_unit, "type": "number"})
        return chunk

    def default_value(self):
        """ Handle measures default values defined as floats, Measure object and force a value when undefined. """
        if self.has_default_value:
            if isinstance(self.attribute.default_value, float):
                default_measure = Measure(self.attribute.default_value)
            elif isinstance(self.attribute.default_value, dict):
                default_measure = Measure(self.attribute.default_value["value"])
            else:
                raise ValueError(f"Default value for attribute '{self.attribute.name}' "
                                 f"must be of type 'float' or 'Measure'. Got '{type(self.attribute.default_value)}'")
            return default_measure.to_dict()
        return {"object_class": f"{Measure.__module__}.{Measure.__class__.__name__}", "value": None}


File = Union[StringFile, BinaryFile]


class FileProperty(Property):
    """ Schema class for File type hints. """

    def __init__(self, annotation: Type[File], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

    @cached_property
    def is_file_related(self) -> bool:
        """ Wether the property is related to files. Base property always is. """
        return True

    @cached_property
    def serialized(self) -> str:
        """ Return file classname. """
        return full_classname(object_=self.annotation, compute_for="class")

    @cached_property
    def pretty_annotation(self) -> str:
        """ Return file classname. """
        return self.annotation.__name__

    def to_dict(self) -> Dict[str, Any]:
        """ Write File as a Dict. """
        chunk = super().to_dict()
        chunk.update({"type": "text", "is_file": True})
        return chunk

    def check_list(self) -> CheckList:
        """
        Check validity of File Type Hint.

        Checks performed :
        - Doesn't define any default value.
        """
        issues = super().check_list()
        issues += CheckList([self.has_no_default()])
        return issues

    def has_no_default(self) -> PassedCheck:
        """ Check if the user definition doesn't have any default value, as it is not supported for files. """
        if self.default_value() is not UNDEFINED:
            msg = f"{self.check_prefix}File input defines a default value, whereas it is not supported."
            return CheckWarning(msg)
        msg = f"{self.check_prefix}File input doesn't define a default value, as it should."
        return PassedCheck(msg)


class CustomClass(Property):
    """ Schema class for CustomClass type hints. """

    def __init__(self, annotation: Type[CoreDessiaObject], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

    @property
    def schema(self) -> ClassSchema:
        """ Return a reference to the schema of the annotation. """
        return ClassSchema(self.annotation)

    @cached_property
    def serialized(self) -> str:
        """ Full class name. """
        return full_classname(object_=self.annotation, compute_for="class")

    @cached_property
    def pretty_annotation(self) -> str:
        """ Return file classname. """
        return self.annotation.__name__

    @property
    def standalone_in_db(self) -> bool:
        """ Whether the class is standalone in db. """
        return self.annotation._standalone_in_db

    def to_dict(self) -> Dict[str, Any]:
        """ Write CustomClass as a Dict. """
        chunk = super().to_dict()
        chunk.update({"type": "object", "standalone_in_db": self.standalone_in_db, "classes": [self.serialized]})
        return chunk

    def default_value(self) -> Optional[Dict[str, Any]]:
        """ Default value for an object. """
        return object_default(default_value=self.attribute.default_value, class_schema=self.schema)

    def check_list(self) -> CheckList:
        """
        Check validity of user custom class Type Hint.

        Checks performed :
        - Is subclass of DessiaObject
        """
        issues = super().check_list()
        issues += CheckList([self.is_dessia_object_typed()])
        return issues

    def is_dessia_object_typed(self) -> PassedCheck:
        """ Check whether if typing for given attribute annotates a subclass of DessiaObject or not . """
        if not issubclass(self.annotation, CoreDessiaObject):
            return WrongType(f"{self.check_prefix}Class '{self.serialized}' is not a subclass of DessiaObject.")
        msg = f"{self.check_prefix}Class '{self.serialized}' is properly typed as a subclass of DessiaObject."
        return PassedCheck(msg)


class UnionProperty(TypingProperty):
    """ Schema class for Union type hints. """

    def __init__(self, annotation: Type[Union[T]], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

    @cached_property
    def serialized(self) -> str:
        """ Generic serialization with 'Union' enforced, because Union annotation has no __name__ attribute. """
        return compute_typing_schema_serialization(serialized_typing="Union", args_schemas=self.args_schemas)

    @cached_property
    def pretty_annotation(self) -> str:
        """ Generic prettifying with 'Union' enforced, because Union annotation has no __name__ attribute. """
        return compute_pretty_schema_annotation(serialized_typing="Union", args_schemas=self.args_schemas)

    @property
    def standalone_in_db(self) -> Optional[bool]:
        """ True if all subclasses are standalone, False if None of them are, None else. """
        standalone_args = [a._standalone_in_db for a in self.args]
        if all(standalone_args):
            return True
        if not any(standalone_args):
            return False
        return None

    @classmethod
    def annotation_from_serialized(cls, serialized: str) -> Type[Union[T]]:
        """ Deserialize Union annotation. """
        return Union[TypingProperty._args_from_serialized(serialized)]

    @property
    def classes(self) -> List[str]:
        """ Compute all possible classes for this annotation. Every class specified in Union annotation. """
        return [full_classname(object_=a, compute_for="class") for a in self.args]

    def to_dict(self) -> Dict[str, Any]:
        """ Write Union as a Dict. """
        chunk = super().to_dict()
        chunk.update({"type": "object", "classes": self.classes, "standalone_in_db": self.standalone_in_db})
        return chunk

    def default_value(self) -> T:
        """ Default value for an object. """
        return object_default(default_value=self.attribute.default_value)

    def check_list(self) -> CheckList:
        """
        Check validity of UnionProperty Type Hint.

        Checks performed :
        - Sub-objects are all standalone or none of them are. TODO : What happen if arguments are not DessiaObjects ?
        """
        issues = super().check_list()
        issues += CheckList([self.classes_are_standalone_consistent()])
        return issues

    def classes_are_standalone_consistent(self) -> PassedCheck:
        """ Check whether all class in Union are standalone or none of them are. """
        standalone_args = [a._standalone_in_db for a in self.args]
        if all(standalone_args):
            msg = f"{self.check_prefix}All arguments of Union type '{self.annotation}' are standalone in db."
            return PassedCheck(msg)
        if not any(standalone_args):
            msg = f"{self.check_prefix}No arguments of Union type '{self.annotation}' are standalone in db."
            return PassedCheck(msg)
        msg = f"{self.check_prefix}'standalone_in_db' values for arguments of Union type '{self.annotation}'" \
              f"are not consistent. They should be all standalone in db or none of them should."
        return WrongType(msg)


class HeterogeneousSequence(TypingProperty):
    """
    Schema class for Tuple type hints.

    Datatype that can be seen as a Tuple. Have any amount of arguments but a limited length.
    """

    def __init__(self, annotation: Type[Tuple], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

        self.additional_items = Ellipsis in self.args

    @property
    def args_schemas(self) -> List[Property]:
        """
        If length is undefined (additional_items is True), then we only have one possible argument type.

        Otherwise, each argument is ordered and strictly defined by its type
        """
        if self.additional_items:
            subattribute = SchemaAttribute(name=f"{self.attribute.name}/0", title=f"{self.attribute.title}/0",
                                           editable=self.attribute.editable)
            return [get_schema(annotation=self.args[0], attribute=subattribute)]
        schemas = []
        for i, arg in enumerate(self.args):
            subattribute = SchemaAttribute(name=f"{self.attribute.name}/{i}", title=f"{self.attribute.title}/{i}",
                                           editable=self.attribute.editable)
            schemas.append(get_schema(annotation=arg, attribute=subattribute))
        return schemas

    @cached_property
    def serialized(self) -> str:
        """ If additional items, concatenate ellipsis. """
        if self.additional_items:
            return f"Tuple[{self.args_schemas[0].serialized}, ...]"
        return super().serialized

    @cached_property
    def pretty_annotation(self) -> str:
        """ If additional items, concatenate ellipsis. """
        if self.additional_items:
            return f"Tuple[{self.args_schemas[0].pretty_annotation}, ...]"
        return super().pretty_annotation

    @classmethod
    def annotation_from_serialized(cls, serialized: str) -> Type[Tuple]:
        """ Deserialize Tuple annotation with undefined length support. """
        rawargs = cls._raw_args_from_serialized(serialized)
        if ",..." in rawargs:
            arg = rawargs.split(",...")[0]
            subtype = deserialize_annotation(arg)
            return Tuple[subtype, ...]
        return Tuple[TypingProperty._args_from_serialized(serialized)]

    def to_dict(self) -> Dict[str, Any]:
        """ Write HeterogeneousSequence as a Dict. """
        chunk = super().to_dict()
        items = []
        for i, schema in enumerate(self.args_schemas):
            schema.title = f"{self.attribute.title}/{i}"
            schema.editable = self.attribute.editable
            items.append(schema.to_dict())
        chunk.update({"type": "array", "additionalItems": self.additional_items, "items": items})
        return chunk

    def default_value(self) -> Tuple[T, ...]:
        """
        Default value for a Tuple.

        Return serialized user default if defined, else a right-sized Tuple filled with None values .
        """
        if self.has_default_value:
            return self.attribute.default_value
        return tuple(s.default_value() for s in self.args_schemas)

    def check_list(self) -> CheckList:
        """ Check validity of Tuple Type Hint. """
        issues = super().check_list()
        issues += CheckList([self.has_enough_args(), self.ellipsis_has_exactly_two_args()])
        return issues

    def has_enough_args(self) -> PassedCheck:
        """ Annotation should have at least one argument, one for each element of the Tuple. """
        if len(self.args) == 0:
            msg = f"{self.check_prefix}is typed as a 'Tuple' which requires at least 1 argument. " \
                  f"Expected 'Tuple[T0, T1, ..., Tn]', got '{self.annotation}'."
            return WrongNumberOfArguments(msg)
        return PassedCheck(f"{self.check_prefix}has at least one argument : '{self.annotation}'.")

    def ellipsis_has_exactly_two_args(self) -> PassedCheck:
        """
        Tuple can use Ellipsis (Tuple[T, ...]), meaning that it contains any number of element.

        In this case it MUST have exactly two arguments.
        """
        if self.additional_items and len(self.args) != 2:
            msg = f"{self.check_prefix}is typed as an ellipsed 'Tuple' which requires at exactly 2 arguments. " \
                  f"Expected 'Tuple[T, ...]', got '{self.annotation}'."
            return WrongNumberOfArguments(msg)
        return PassedCheck(f"{self.check_prefix}is not an ill-defined ellipsed tuple : '{self.annotation}'.")


class HomogeneousSequence(TypingProperty):
    """
    Schema class for List type hints.

    Datatype that can be seen as a List. Have only one argument but an unlimited length.
    """

    def __init__(self, annotation: Type[List[T]], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

    @classmethod
    def annotation_from_serialized(cls, serialized: str) -> Type[List]:
        """ Deserialize List annotation. """
        return List[TypingProperty._args_from_serialized(serialized)]

    def to_dict(self) -> Dict[str, Any]:
        """ Write HomogeneousSequence as a Dict. """
        if self.attribute.title:
            title = self.attribute.title
        else:
            title = "Items"
        chunk = super().to_dict()
        schema = self.args_schemas[0]
        schema.title = f"{title} items"
        schema.editable = self.attribute.editable
        items = schema.to_dict()
        chunk.update({"type": "array", "items": items})
        return chunk

    def default_value(self) -> Type[None]:
        """ Default of a sequence. Always return None as default mutable is prohibited. """
        return None

    def check_list(self) -> CheckList:
        """ Check validity of List Type Hint. """
        issues = super().check_list()
        issues += CheckList([self.has_one_arg(), self.has_no_default()])
        return issues

    def has_no_default(self) -> PassedCheck:
        """ Check if List doesn't define a default value that is other than None. """
        if self.default_value() is not UNDEFINED and self.default_value() is not None:
            msg = f"{self.check_prefix}Mutable List input defines a default value other than None," \
                  f"which will lead to unexpected behavior and therefore, is not supported."
            return UnsupportedDefault(msg)
        msg = f"{self.check_prefix}Mutable List doesn't define a default value other than None."
        return PassedCheck(msg)


class DynamicDict(TypingProperty):
    """
    Schema class for Dict type hints.

    Datatype that can be seen as a Dict. Have restricted amount of arguments (one for key, one for values),
    but an unlimited length.
    """

    def __init__(self, annotation: Type[Dict[str, Builtin]], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

    @classmethod
    def annotation_from_serialized(cls, serialized: str) -> Type[Dict]:
        """ Deserialize Dict annotation. """
        return Dict[TypingProperty._args_from_serialized(serialized)]

    def to_dict(self) -> Dict[str, Any]:
        """ Write DynamicDict as a Dict. """
        key_type, value_type = self.args
        if key_type != str:
            # !!! Should we support other types ? Numeric ?
            raise NotImplementedError("Non strings keys not supported")
        if value_type not in TYPING_EQUIVALENCES:
            raise ValueError(f"Dicts should have only builtins keys and values, got {value_type}")
        chunk = super().to_dict()
        chunk.update({"type": "object",
                      "patternProperties": {
                          ".*": {
                            "type": TYPING_EQUIVALENCES[value_type]
                          }
                      }})
        return chunk

    def default_value(self) -> Type[None]:
        """ Default of a dynamic Dict. Always return None as default mutable is prohibited. """
        return None

    def check_list(self) -> CheckList:
        """ Check validity of DynamicDict Type Hint. """
        issues = super().check_list()
        checks = [self.has_two_args(), self.has_string_keys(), self.has_simple_values(), self.has_no_default()]
        issues += CheckList(checks)
        return issues

    def has_two_args(self) -> PassedCheck:
        """ Annotation should have exactly two arguments, first one for keys, second one for values. """
        if len(self.args) != 2:
            msg = f"{self.check_prefix}is typed as a 'Dict' which requires exactly 2 arguments. " \
                  f"Expected 'Dict[KeyType, ValueType]', got '{self.annotation}'."
            return WrongNumberOfArguments(msg)
        return PassedCheck(f"{self.check_prefix}has two args in its definition : '{self.annotation}'.")

    def has_string_keys(self) -> PassedCheck:
        """ Key Type should be str. """
        key_type, value_type = self.args
        if not issubclass(key_type, str):
            # Should we support other types ? Numeric ?
            msg = f"{self.check_prefix}is typed as a 'Dict[{key_type}, {value_type}]' " \
                  f"which requires str as its key type. Expected 'Dict[str, ValueType]', got '{self.annotation}'."
            return WrongType(msg)
        return PassedCheck(f"{self.check_prefix}has str keys : '{self.annotation}'.")

    def has_simple_values(self) -> PassedCheck:
        """ Value Type should be simple. """
        key_type, value_type = self.args
        if value_type not in TYPING_EQUIVALENCES:
            msg = f"{self.check_prefix}is typed as a 'Dict[{key_type}, {value_type}]' " \
                  f"which requires a builtin type as its value type. " \
                  f"Expected 'int', 'float', 'bool' or 'str', got '{value_type}'."
            return WrongType(msg)
        return PassedCheck(f"{self.check_prefix}has simple values : '{self.annotation}'.")

    def has_no_default(self) -> PassedCheck:
        """ Check if Dict doesn't define a default value that is other than None. """
        if self.default_value() is not UNDEFINED and self.default_value() is not None:
            msg = f"{self.check_prefix}Mutable Dict input defines a default value other than None," \
                  f"which will lead to unexpected behavior and therefore, is not supported."
            return UnsupportedDefault(msg)
        msg = f"{self.check_prefix}Mutable Dict doesn't define a default value other than None."
        return PassedCheck(msg)


BaseClass = TypeVar("BaseClass", bound=CoreDessiaObject)


class InstanceOfProperty(TypingProperty):
    """
    Schema class for InstanceOf type hints.

    Datatype that can be seen as a union of classes that inherits from the only argument given.
    Instances of these classes validate against this type.
    """

    def __init__(self, annotation: Type[InstanceOf[BaseClass]], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

    @classmethod
    def annotation_from_serialized(cls, serialized: str) -> Type[InstanceOf[BaseClass]]:
        """ Deserialize InstanceOf annotation. """
        return InstanceOf[TypingProperty._args_from_serialized(serialized)]

    @property
    def schema(self) -> ClassSchema:
        """ Get Schema of base class. """
        return ClassSchema(self.args[0])

    @property
    def standalone_in_db(self) -> bool:
        """ Return True if base class is standalone. """
        return self.args[0]._standalone_in_db

    @property
    def classes(self) -> List[str]:
        """ Compute all possible classes for this annotation. Return base class. """
        return [full_classname(object_=self.args[0], compute_for="class")]

    def to_dict(self) -> Dict[str, Any]:
        """ Write InstanceOf as a Dict. """
        chunk = super().to_dict()
        chunk.update({"type": "object", "instance_of": self.classes[0], "classes": self.classes,
                      "standalone_in_db": self.standalone_in_db})
        return chunk

    def default_value(self) -> Dict[str, Any]:
        """ Default value of an object. """
        return object_default(default_value=self.attribute.default_value, class_schema=self.schema)

    def check_list(self) -> CheckList:
        """ Check validity of InstanceOf Type Hint. """
        issues = super().check_list()
        issues += CheckList([self.has_one_arg()])
        return issues


class SubclassProperty(TypingProperty):
    """
    Schema class for Subclass type hints.

    Datatype that can be seen as a union of classes that inherits from the only argument given.
    Classes validate against this type.
    """

    def __init__(self, annotation: Type[Subclass[BaseClass]], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

    @classmethod
    def annotation_from_serialized(cls, serialized: str) -> Type[Subclass[BaseClass]]:
        """ Deserialize Subclass annotation. """
        return Subclass[TypingProperty._args_from_serialized(serialized)]

    @property
    def standalone_in_db(self) -> bool:
        """ Return True if base class is standalone. """
        return self.args[0]._standalone_in_db

    def to_dict(self) -> Dict[str, Any]:
        """ Write Subclass as a Dict. """
        raise NotImplementedError("Subclass is not implemented yet")

    def check_list(self) -> CheckList:
        """
        Check validity of Subclass Type Hint.

        Checks performed :
        - Annotation has exactly one argument, which is the type of the base class.
        """
        issues = super().check_list()
        issues += CheckList([self.has_one_arg()])
        return issues


class AttributeTypeProperty(TypingProperty):
    """
    Schema class for AttributeType and ClassAttributeType type hints.

    A specifically instantiated AttributeType validated against this type.
    """

    def __init__(self, annotation: Type[AttributeType], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

        self.class_ = self.args[0]
        if self.class_ is Type:
            subdefault = UNDEFINED
        else:
            subdefault = self.class_
        subattribute = SchemaAttribute(name=f"{attribute.name}/class", default_value=subdefault, title="Class",
                                       editable=attribute.editable)
        self.class_schema = get_schema(annotation=Type[self.class_], attribute=subattribute)

    @classmethod
    def annotation_from_serialized(cls, serialized: str) -> Type[AttributeType]:
        """ Deserialize Attribute annotation. Support Class and Instance attributes. """
        type_ = TypingProperty.type_from_serialized(serialized)
        if type_ == "AttributeType":
            return AttributeType[TypingProperty._args_from_serialized(serialized)]
        return ClassAttributeType[TypingProperty._args_from_serialized(serialized)]

    def to_dict(self) -> Dict[str, Any]:
        """ Write AttributeType as a Dict. """
        chunk = super().to_dict()
        if self.origin is ClassAttributeType:
            attribute_type = "class_attributes"
        else:
            attribute_type = "attributes"
        classname = full_classname(self.origin, compute_for="class")
        chunk.update({
            "type": "object", "object_class": classname, "is_attribute": True,
            "attribute_type": attribute_type,
            "properties": {
                "class_": self.class_schema.to_dict(),
                "name": {
                    "type": "string"
                }
            }
        })
        return chunk
    
    def default_value(self) -> Dict[str, Any]:
        """ Sets AttributeType object_class and argument class_ if it is different from Type. """
        if self.has_default_value:
            return self.attribute.default_value.to_dict()
        
        if self.class_ is not Type and issubclass(self.class_, CoreDessiaObject):
            classname = full_classname(object_=self.class_, compute_for="class")
        else:
            classname = None
        return {"object_class": full_classname(object_=self.origin, compute_for="class"),
                "class_": classname, "name": None}

    def check_list(self) -> CheckList:
        """
        Check validity of MethodType Type Hint.

        Checks performed :
        - Class has method TODO
        """
        return CheckList([])


class MethodTypeProperty(AttributeTypeProperty):
    """
    Schema class for MethodType and ClassMethodType type hints.

    A specifically instantiated MethodType validated against this type.
    """

    def __init__(self, annotation: Type[MethodType], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

    @classmethod
    def annotation_from_serialized(cls, serialized: str) -> Type[MethodType]:
        """ Deserialize Methods annotation. Support Class and Instance methods. """
        type_ = TypingProperty.type_from_serialized(serialized)
        if type_ == "MethodType":
            return MethodType[TypingProperty._args_from_serialized(serialized)]
        return ClassMethodType[TypingProperty._args_from_serialized(serialized)]

    def to_dict(self) -> Dict[str, Any]:
        """ Write MethodType as a Dict. """
        chunk = super().to_dict()
        if self.origin is ClassMethodType:
            attribute_type = "class_methods"
        else:
            attribute_type = "methods"
        chunk["attribute_type"] = attribute_type
        return chunk


class SelectorProperty(AttributeTypeProperty):
    """
    Schema class for AttributeType and ClassAttributeType type hints.

    A specifically instantiated AttributeType validated against this type.
    """

    def __init__(self, annotation: Type[ViewType], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

    @classmethod
    def annotation_from_serialized(cls, serialized: str) -> Type[ViewType]:
        """ Deserialize Attribute annotation. Support Class and Instance attributes. """
        type_ = TypingProperty.type_from_serialized(serialized)
        if type_ == "CadViewType":
            return CadViewType[TypingProperty._args_from_serialized(serialized)]
        if type_ == "MarkdownType":
            return MarkdownType[TypingProperty._args_from_serialized(serialized)]
        if type_ == "PlotDataType":
            return PlotDataType[TypingProperty._args_from_serialized(serialized)]
        raise NotImplementedError("Other Display than CadView not implemented yet")

    def to_dict(self) -> Dict[str, Any]:
        """ Write AttributeType as a Dict. """
        chunk = super().to_dict()
        chunk.update({"attribute_type": "view_selectors", "decorator": self.annotation.decorator})
        return chunk


Class = TypeVar("Class", bound=type)


class ClassProperty(TypingProperty):
    """
    Schema class for 'Type' type hints.

    Non DessiaObject sub-classes validated against this type.
    """

    def __init__(self, annotation: Type[Class], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

        if annotation is Type:
            self.class_ = Type
        else:
            self.class_ = self.args[0]

    @classmethod
    def annotation_from_serialized(cls, serialized: str) -> Type[Class]:
        """ Deserialize Type annotation. Support undefined and defined argument. """
        args = TypingProperty._args_from_serialized(serialized)
        if args:
            return Type[args]
        return Type

    def to_dict(self) -> Dict[str, Any]:
        """ Write Class as a Dict. """
        chunk = super().to_dict()
        chunk.update({"type": "object", "is_class": True, "properties": {"name": {"type": "string"}}})
        return chunk

    def default_value(self) -> str:
        """ Sets AttributeType object_class and argument class_ if it is different from Type. """
        if self.has_default_value:
            return full_classname(self.attribute.default_value, compute_for="class")

        if self.class_ is not Type and issubclass(self.class_, CoreDessiaObject):
            return full_classname(object_=self.class_, compute_for="class")
        return ""

    def check_list(self) -> CheckList:
        """
        Check validity of Class Type Hint.

        Checks performed :
        - Annotation has exactly 1 argument
        """
        issues = super().check_list()
        issues += CheckList([self.has_one_arg()])
        return issues


class GenericTypeProperty(Property):
    """ Meta Property for Types. """

    def __init__(self, annotation: Type[TypeVar], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

    @classmethod
    def annotation_from_serialized(cls, serialized: str) -> Type[TypeVar]:
        """ Deserialize Generic Type annotation. """
        raise NotImplementedError("Annotation deserialization not implemented for Generic Types")

    @cached_property
    def serialized(self) -> Optional[str]:
        """ Serialize None as itself as json will dump it as null. Otherwise, serialize as typing.Type. """
        if self.annotation is None:
            return None
        return "typing.Type"

    def to_dict(self) -> Dict[str, Any]:
        """ Write Type as a Dict. """
        chunk = super().to_dict()
        chunk.update({"type": "object", "is_class": True, "properties": {"name": {"type": "string"}}})
        return chunk


class AnyProperty(Property):
    """ Handle Any typed (cannot be form inputs). """

    def __init__(self, annotation: Type[T], attribute: SchemaAttribute):
        super().__init__(annotation=annotation, attribute=attribute)

    def to_dict(self) -> Dict[str, Any]:
        """ Write chunk of Any property. Useful for low-code features. """
        chunk = super().to_dict()
        chunk.update({"properties": {".*": ".*"}, "type": "object"})
        return chunk


def inspect_arguments(method: Callable, merge: bool = False) -> Tuple[List[str], Dict[str, Any]]:
    """ Wrapper around 'split_default_argument' method in order to call it from a method object. """
    method_full_name = f"{method.__module__}.{method.__qualname__}"
    if method_full_name in _fullargsspec_cache:
        argspecs = _fullargsspec_cache[method_full_name]
    else:
        argspecs = inspect.getfullargspec(method)
        _fullargsspec_cache[method_full_name] = argspecs
    return split_default_args(argspecs=argspecs, merge=merge)


def split_default_args(argspecs: inspect.FullArgSpec, merge: bool = False) -> Tuple[List[str], Dict[str, Any]]:
    """
    Find default value and required arguments of class construction.

    Get method arguments and default arguments as sequences while removing forbidden ones (self, cls...).
    """
    nargs, ndefault_args = split_argspecs(argspecs)

    default_arguments = {}
    arguments = []
    for iargument, argument in enumerate(argspecs.args[1:]):
        if argument not in RESERVED_ARGNAMES:
            if iargument >= nargs - ndefault_args:
                default_value = argspecs.defaults[ndefault_args - nargs + iargument]
                if merge:
                    arguments.append((argument, default_value))
                else:
                    default_arguments[argument] = default_value
            else:
                arguments.append(argument)
    return arguments, default_arguments


def split_argspecs(argspecs: inspect.FullArgSpec) -> Tuple[int, int]:
    """ Get number of regular arguments as well as arguments with default values. """
    nargs = len(argspecs.args) - 1
    if argspecs.defaults is not None:
        ndefault_args = len(argspecs.defaults)
    else:
        ndefault_args = 0
    return nargs, ndefault_args


def get_schema(annotation: Type[T], attribute: SchemaAttribute) -> Property:
    """ Get schema Property object from given annotation. """
    if annotation is None or inspect.isclass(annotation) and issubclass(annotation, type(None)):
        schema_type = GenericTypeProperty
    elif annotation in TYPING_EQUIVALENCES:
        schema_type = BuiltinProperty
    elif is_typing(annotation):
        return typing_schema(typing_=annotation, attribute=attribute)
    elif hasattr(annotation, "__origin__") and annotation.__origin__ is type:
        # Type is not considered a Typing as it has no arguments
        schema_type = ClassProperty
    elif annotation is Any:
        schema_type = AnyProperty
    elif inspect.isclass(annotation):
        return custom_class_schema(annotation=annotation, attribute=attribute)
    elif isinstance(annotation, TypeVar):
        schema_type = GenericTypeProperty
    else:
        raise NotImplementedError(f"No schema defined for attribute '{attribute}' annotated '{annotation}'.")
    return schema_type(annotation=annotation, attribute=attribute)


ORIGIN_TO_SCHEMA_CLASS = {
    tuple: HeterogeneousSequence, list: HomogeneousSequence,
    collections.abc.Iterator: HomogeneousSequence, Union: UnionProperty,
    dict: DynamicDict, InstanceOf: InstanceOfProperty,
    type: ClassProperty,
    MethodType: MethodTypeProperty, ClassMethodType: MethodTypeProperty,
    AttributeType: AttributeTypeProperty, ClassAttributeType: AttributeTypeProperty,
    CadViewType: SelectorProperty, PlotDataType: SelectorProperty, MarkdownType: SelectorProperty
}

SERIALIZED_TO_SCHEMA_CLASS = {
    "int": BuiltinProperty, "float": BuiltinProperty, "bool": BuiltinProperty, "str": BuiltinProperty,
    "Tuple": HeterogeneousSequence, "List": HomogeneousSequence, "Iterator": HomogeneousSequence,
    "Union": UnionProperty, "Dict": DynamicDict, "InstanceOf": InstanceOfProperty, "Subclass": SubclassProperty,
    "MethodType": MethodTypeProperty, "ClassMethodType": MethodTypeProperty, "Type": ClassProperty,
    "AttributeType": AttributeTypeProperty, "ClassAttributeType": AttributeTypeProperty
}


def serialize_annotation(annotation: Type[T], attribute_name: str = "") -> str:
    """ Make use of schema to serialize annotations. """
    attribute = SchemaAttribute(attribute_name)
    schema = get_schema(annotation=annotation, attribute=attribute)
    return schema.serialized


def pretty_annotation(annotation: Type[T], attribute_name: str = "") -> str:
    """ Make use of schema to compute pretty annotations. """
    attribute = SchemaAttribute(attribute_name)
    schema = get_schema(annotation=annotation, attribute=attribute)
    return schema.pretty_annotation


def deserialize_annotation(serialized: str) -> Type[T]:
    """ From a string denoting an annotation, get deserialize value. """
    if serialized is None:
        return None
    if "[" in serialized:
        return TypingProperty.annotation_from_serialized(serialized)
    if serialized in SERIALIZED_TO_SCHEMA_CLASS:
        return SERIALIZED_TO_SCHEMA_CLASS[serialized].annotation_from_serialized(serialized)
    return get_python_class_from_class_name(serialized)


def is_typing(object_) -> bool:
    """ Return True if given object can be seen as a typing (has a module, an origin and arguments). """
    has_module = hasattr(object_, "__module__")
    has_origin = hasattr(object_, "__origin__")
    has_args = hasattr(object_, "__args__")
    return has_module and has_origin and has_args


def typing_schema(typing_: Type[T], attribute: SchemaAttribute) -> Property:
    """ Get schema Property for typing annotations. """
    origin = get_origin(typing_)
    if origin is Union and union_is_default_value(typing_):
        # This is a false UnionProperty => Is a default value set to None
        schema_type = OptionalProperty
    else:
        try:
            schema_type = ORIGIN_TO_SCHEMA_CLASS[origin]
        except KeyError as exc:
            raise NotImplementedError(f"No Schema defined for typing '{typing_}'.") from exc
    return schema_type(annotation=typing_, attribute=attribute)


def custom_class_schema(annotation: Type[T], attribute: SchemaAttribute) -> Property:
    """ Get schema Property object for non typing annotations. """
    if issubclass(annotation, Measure):
        schema_type = MeasureProperty
    elif issubclass(annotation, (BinaryFile, StringFile)):
        schema_type = FileProperty
    elif issubclass(annotation, CoreDessiaObject):
        # Dessia custom classes
        schema_type = CustomClass
    elif issubclass(annotation, DisplayObject):
        schema_type = Property
    else:
        raise NotImplementedError(f"No Schema defined for type '{annotation}'.")
    return schema_type(annotation=annotation, attribute=attribute)


def object_default(default_value: CoreDessiaObject = UNDEFINED, class_schema: ClassSchema = None) \
        -> Optional[Dict[str, Any]]:
    """
    Default value of an object.

    Return serialized user default if definition, else None.
    """
    if default_value is not UNDEFINED:
        return default_value.to_dict(use_pointers=False)
    if class_schema is not None:
        # TODO Should we implement this ? Right now, tests state that the result is None
        # return class_schema.default_dict()
        pass
    return None


def compute_typing_schema_serialization(serialized_typing: str, args_schemas: List[Property]) -> str:
    """ Build final typing serialized string. """
    return f"{serialized_typing}[{', '.join([s.serialized for s in args_schemas])}]"


def compute_pretty_schema_annotation(serialized_typing: str, args_schemas: List[Property]) -> str:
    """ Build final typing pretty annotation string. """
    return f"{serialized_typing}[{', '.join([s.pretty_annotation for s in args_schemas])}]"


def serialize_typing(typing_, attribute: str = "") -> str:
    """ Make use of schema to serialized annotations. """
    warnings.warn("Function serialize_typing have been renamed serialize_annotation. Please use it instead.",
                  DeprecationWarning)
    return serialize_annotation(annotation=typing_, attribute_name=attribute)


def union_is_default_value(typing_: Type) -> bool:
    """
    Union typings can be False positives.

    An argument of a function that has a default_value set to None is Optional[T], which is an alias for
    Union[T, NoneType]. This function checks if this is the case.
    """
    args = get_args(typing_)
    return len(args) == 2 and type(None) in args


def extract_args(string: str) -> List[str]:
    """
    Extract first level arguments from serialized typing.

    This function does not split by commas when we are down to second level annotations
    and has been preferred to regexp for this reason.
    """
    opened_brackets = 0
    closed_brackets = 0
    current_arg = ""
    arguments = []
    for character in string.replace(" ", ""):
        if character == "[":
            opened_brackets += 1
        if character == "]":
            closed_brackets += 1
        split_by_comma = closed_brackets == opened_brackets
        if split_by_comma and character == ",":
            # We are at first level, because we closed as much brackets as we have opened
            # Current argument is complete, we append it to arguments sequence and reset current arguments
            arguments.append(current_arg)
            current_arg = ""
        else:
            # We are at a deeper level, because all opened brackets haven't been closed.
            # We build current argument
            current_arg += character
    if current_arg:
        # Append last argument to arguments sequence
        arguments.append(current_arg)
    return arguments


def parse_class_docstring(class_) -> ParsedDocstring:
    """ Helper to get parse docstring from a class. """
    docstring = class_.__doc__
    annotations = get_type_hints(class_.__init__)
    return parse_docstring(docstring=docstring, annotations=annotations)


def parse_docstring(docstring: str, annotations: Dict[str, Any]) -> ParsedDocstring:
    """ Parse user-defined docstring of given class. Refer to docs to see how docstring should be built. """
    if docstring:
        no_return_docstring = docstring.split(":return:")[0]
        splitted_docstring = no_return_docstring.split(":param ")
        parsed_docstring = {"description": splitted_docstring[0].strip()}
        params = splitted_docstring[1:]
        args = {}
        for param in params:
            argname, parsed_attribute = parse_attribute(param, annotations)
            args[argname] = parsed_attribute
        parsed_docstring.update({"attributes": args})
        return parsed_docstring
    return {"description": "", "attributes": {}}


def parse_attribute(param, annotations) -> Tuple[str, ParsedAttribute]:
    """ Extract attribute from user-defined docstring. """
    if ":type" in param:
        param = param.split(":type ")[0]
    argname, argdesc = param.split(":", maxsplit=1)
    annotation = annotations[argname]
    parsed_attribute = {"desc": argdesc.strip(), "type_": serialize_annotation(annotation),
                        "annotation": str(annotation)}
    return argname, parsed_attribute


def _check_docstring(element):
    """ Return True if an object, a class or a method have a proper docstring. Otherwise, return False. """
    docstring = element.__doc__
    if docstring is None:
        print(f"Docstring not found for {element}")
        return False
    if inspect.isclass(element):
        # element is an object or a class
        annotations = get_type_hints(element.__init__)
    elif inspect.ismethod(element) or inspect.isfunction(element):
        # element is a method
        annotations = get_type_hints(element)
    else:
        raise NotImplementedError
    try:
        parse_docstring(docstring=docstring, annotations=annotations)
        return True
    except TypeError:
        return False



"""
__init__ method for models module
"""
from .cars import all_cars_no_feat, all_cars_wi_feat
from .rand_data import rand_data_small, rand_data_middl, rand_data_large
from .power_test import simulation1, simulation_list, system1
from .workflows.workflow_exports import workflow_export, workflow_export_state



from dessia_common.forms import Generator, StandaloneObject
from dessia_common.workflow.blocks import InstantiateModel, ClassMethod, Sequence
from dessia_common.typings import ClassMethodType
from dessia_common.workflow.core import Variable, Pipe, Workflow
from dessia_common.files import StringFile

block_0 = InstantiateModel(model_class=Generator, name='Simple Inputs',
                           position=(41.40861618798957, -84.52480417754569))
block_1 = ClassMethod(method_type=ClassMethodType(StandaloneObject, 'generate_from_bin'), name='Binary Input',
                      position=(42.59477666877444, 94.63608398070332))
block_2 = ClassMethod(method_type=ClassMethodType(StandaloneObject, 'generate_from_text'), name='Text Input',
                      position=(42.32922270510359, 184.5889177760892))
block_3 = InstantiateModel(model_class=StandaloneObject, name='Custom & Complex Inputs',
                           position=(39.02562760496994, 274.3967730534379))
block_4 = Sequence(number_arguments=4, name='Packer', position=(482.55623557963327, 33.78454658062492))
blocks = [block_0, block_1, block_2, block_3, block_4]

variable_0 = Variable(name='Result Name', position=(-353.5235486758587, 234.9165175396766), type_=str)
variable_1 = Variable(name='stream', position=(-131.55327139280053, 248.81817426839467), type_=StringFile)

pipe_0 = Pipe(variable_0, block_3.inputs[12])
pipe_1 = Pipe(variable_0, block_0.inputs[3])
pipe_2 = Pipe(variable_1, block_2.inputs[0])
pipe_3 = Pipe(block_0.outputs[0], block_4.inputs[0])
pipe_4 = Pipe(block_1.outputs[0], block_4.inputs[1])
pipe_5 = Pipe(block_2.outputs[0], block_4.inputs[2])
pipe_6 = Pipe(block_3.outputs[0], block_4.inputs[3])
pipes = [pipe_0, pipe_1, pipe_2, pipe_3, pipe_4, pipe_5, pipe_6]

workflow = Workflow(blocks, pipes, output=block_4.outputs[0], name='Test Inputs')

assert workflow.has_file_inputs
assert len(workflow.file_inputs) == 2

file_input_indices = [workflow.input_index(i) for i in workflow.file_inputs]

assert file_input_indices == [3, 17]
assert workflow.is_variable_nbv(workflow.file_inputs[1])



""" __init__ method for datatools module """

import warnings
from typing import List
from dessia_common.core import DessiaObject
import dessia_common.datatools.dataset as DS
import dessia_common.datatools.cluster as DC

# Imports for retro-compatibility
# TODO: Remove it kindly in next releases.


class HeterogeneousList(DS.Dataset):
    def __init__(self, dessia_objects: List[DessiaObject] = None, name: str = ''):
        self.warning_string()
        DS.Dataset.__init__(self, dessia_objects=dessia_objects, name=name)

    def warning_string(self):
        string = "Class HeterogeneousList is not supported anymore and will be deleted in next releases (0.11.0).\n"
        string += "Please use the Dataset class, which is exactly the same.\n"
        string += "Dataset is imported with <from dessia_common.datatools.dataset import Dataset>.\n"
        warnings.warn(string, DeprecationWarning)


class CategorizedList(DC.ClusteredDataset):
    def __init__(self, dessia_objects: List[DessiaObject] = None, labels: List[int] = None, name: str = ''):
        self.warning_string()
        DC.ClusteredDataset.__init__(self, dessia_objects=dessia_objects, labels=labels, name=name)

    def warning_string(self):
        string = "Class CategorizedList is not supported anymore and will be deleted in next releases (0.11.0).\n"
        string += "Please use the ClusteredDataset class, which is exactly the same.\n"
        string += "ClusteredDataset is imported with <from dessia_common.datatools.cluster import ClusteredDataset>.\n"
        warnings.warn(string, DeprecationWarning)



""" Abstract module that defines a base DessiaObject in order to avoid circular imports. """


ABSTRACT_ERROR = NotImplementedError("CoreDessiaObject is an abstract class and should not be use directly.")


class CoreDessiaObject:
    """ Base DessiaObject for checking inheritance purpose. """

    def to_dict(self, use_pointers: bool = True, memo=None, path: str = "#", id_method: bool = True, id_memo=None):
        """ Abstract to_dict method. """
        raise ABSTRACT_ERROR

    @classmethod
    def dict_to_object(cls, dict_, **kwargs):
        """ Abstract dict_to_object method. """
        raise ABSTRACT_ERROR

    def _data_diff(self, other_object: 'CoreDessiaObject'):
        """ Abstract _data_diff method. """
        raise ABSTRACT_ERROR

    @property
    def method_schemas(self):
        """ Abstract method_schemas method. """
        raise ABSTRACT_ERROR

    def _displays(self):
        """ Abstract _displays method. """
        raise ABSTRACT_ERROR



"""
Tests for dessia_common.DessiaFilters/FiltersList class
"""
import itertools
from dessia_common.models import all_cars_no_feat, all_cars_wi_feat, rand_data_large
from dessia_common.core import DessiaFilter, FiltersList
from dessia_common.datatools.dataset import Dataset

# When attribute _features is not specified in class Car
all_cars_without_features = Dataset(all_cars_no_feat)
# When attribute _features is specified in class CarWithFeatures
all_cars_with_features = Dataset(all_cars_wi_feat)
# Auto-generated heterogeneous dataset with nb_clusters clusters of points in nb_dims dimensions
RandData_heterogeneous = Dataset(rand_data_large)

# Filters creation
weight_val = 2000.
mpg_big_val = 100.
mpg_low_val = 40.
filter_1 = DessiaFilter('weight', 'le', weight_val)
filter_2 = DessiaFilter('mpg', 'ge', mpg_big_val)
filter_3 = DessiaFilter('mpg', 'ge', mpg_low_val)
filters = [filter_1, filter_2, filter_3]
filters_list = FiltersList(filters, logical_operator="or")
print(filters_list)

# Or testing
filters_list_fun = lambda x: ((getattr(value, 'weight') <= weight_val or
                               getattr(value, 'mpg') >= mpg_big_val or
                               getattr(value, 'mpg') >= mpg_low_val)
                              for value in all_cars_no_feat)

assert(all(item in all_cars_without_features.filtering(filters_list)
           for item in list(itertools.compress(all_cars_no_feat, filters_list_fun(all_cars_no_feat)))))

# And with non empty result
filters_list = FiltersList([filter_1, filter_3], logical_operator="and")
filters_list_fun = lambda x: ((getattr(value, 'weight') <= weight_val and getattr(value, 'mpg') >= mpg_low_val)
                              for value in all_cars_no_feat)
assert(all(item in all_cars_without_features.filtering(filters_list)
           for item in list(itertools.compress(all_cars_no_feat, filters_list_fun(all_cars_no_feat)))))

# And with empty result
filters_list = FiltersList([filter_1, filter_2], logical_operator="and")
assert(all_cars_without_features.filtering(filters_list) == Dataset())

# Xor
filter_1 = DessiaFilter('weight', 'le', weight_val)
filter_3 = DessiaFilter('mpg', 'ge', mpg_low_val)
filters_list = FiltersList([filter_1, filter_2, filter_3], logical_operator="xor")
filters_list_fun = lambda x: ((getattr(value, 'weight') <= weight_val
                               and not getattr(value, 'mpg') >= mpg_big_val
                               and not getattr(value, 'mpg') >= mpg_low_val) or
                              (not getattr(value, 'weight') <= weight_val
                               and getattr(value, 'mpg') >= mpg_big_val
                               and not getattr(value, 'mpg') >= mpg_low_val) or
                              (not getattr(value, 'weight') <= weight_val
                               and not getattr(value, 'mpg') >= mpg_big_val
                               and getattr(value, 'mpg') >= mpg_low_val)
                          for value in all_cars_no_feat)
assert(all(item in all_cars_without_features.filtering(filters_list)
           for item in list(itertools.compress(all_cars_no_feat, filters_list_fun(all_cars_no_feat)))))

filters_list = FiltersList([filter_1, filter_2, filter_3], logical_operator="blurps")
try:
    all_cars_without_features.filtering(filters_list)
    raise ValueError("'blurps' should not work for logical_operator attribute in FiltersList")
except Exception as e:
    assert(e.args[0] == "'blurps' str for 'logical_operator' attribute is not a use case")




from dessia_common.tests import RandDataD2
from dessia_common.datatools.sampling import ClassSampler
from dessia_common.optimization import FixedAttributeValue, BoundedAttributeValue
from dessia_common.workflow.blocks import InstantiateModel, ModelMethod
from dessia_common.typings import MethodType
from dessia_common.workflow.core import Workflow, Pipe

block_0 = InstantiateModel(model_class=ClassSampler, name='ClassSampler')
block_1 = ModelMethod(method_type=MethodType(ClassSampler, 'make_doe'), name='Make DOE method')
blocks = [block_0, block_1]

pipe_0 = Pipe(block_0.outputs[0], block_1.inputs[0])
pipes = [pipe_0]

workflow = Workflow(blocks, pipes, output=block_1.outputs[0], name='')

sampled_attributes = [BoundedAttributeValue('p_1', 0.1, 0.5, 250)]
constant_attributes = [FixedAttributeValue('p_2', 25)]

# Workflow run
workflow_run = workflow.run({workflow.index(block_0.inputs[0]): RandDataD2,
                             workflow.index(block_0.inputs[1]): sampled_attributes,
                             workflow.index(block_0.inputs[2]): constant_attributes,
                             workflow.index(block_1.inputs[1]): 400})

# Workflow tests
workflow._check_platform()
wfrun_plot_data = workflow_run.output_value._check_platform()

print("script 'workflow_sampling.py' has passed.")



"""
Tests for dessia_common.optimization package
"""
from math import pi, exp, sin
from itertools import product
import matplotlib.pyplot as plt
from typing import List

from dessia_common.core import DessiaObject
import dessia_common.optimization as opt


class Engine(DessiaObject):
    """
    Dummy and unrealistic engine, only for tests on optimization package. Do not use to optimize an engine.
    """
    _vector_features = ['diameter', 'stroke', 'power', 'mass', 'costs']

    def __init__(self, n_cyl: int, diameter: float, stroke: float, r_pow_cyl: float = 1., r_diam_strok: float = 1.,
                 name: str = ''):
        DessiaObject.__init__(self, name=name)
        self.n_cyl = n_cyl
        self.diameter = diameter
        self.stroke = stroke
        self.r_pow_cyl = r_pow_cyl
        self.r_diam_strok = r_diam_strok
        self.cyl_volume = self.compute_cyl_volume()
        self.power = self._power()
        self.mass = self._mass()
        self.costs = self._costs()

    def compute_cyl_volume(self):
        return (self.diameter / 2)**2 * pi * self.stroke

    def cylinder_power(self):
        return (self.cyl_volume * self.r_pow_cyl *
                abs(sin(
                    (self.r_diam_strok * (1 + 0.5 * (self.stroke / self.diameter + self.diameter / self.stroke)))**2)))

    def _power(self):
        return self.n_cyl * self.cylinder_power()

    def carter_volume(self):
        return self.stroke * 1.5 * (1 - exp(-self.diameter)) * 3. * self.diameter * 1.2 * self.n_cyl

    def _mass(self):
        return 7800 * (self.carter_volume() - self.n_cyl * self.cyl_volume)

    def _costs(self):
        return (400 * ((0.175 - self.stroke)**2 + 2 * (0.07 - self.diameter)**2) +
                (sin(self.power / self.r_pow_cyl) + sin(self.mass / 10))**2)

    def to_vector(self):
        return [self.diameter, self.stroke, self.power, self.mass, self.costs]


class EngineOptimizer(opt.InstantiatingModelOptimizer):
    """
    Optimizer for Engine.
    """

    def __init__(self, fixed_parameters: List[opt.FixedAttributeValue],
                 optimization_bounds: List[opt.BoundedAttributeValue],
                 name: str = ''):
        opt.InstantiatingModelOptimizer.__init__(self, fixed_parameters, optimization_bounds, name)

    def instantiate_model(self, attributes_values):
        return Engine(**attributes_values)

    def objective_from_model(self, model, clearance: float = 0.003):
        return model.costs


def check_costs_function(cylinders, diameters, strokes, r_pow_cyl, r_diam_strok):
    points = []
    engines = []
    for n_cyl, diameter, stroke in product(cylinders, diameters, strokes):
        engines.append(Engine(n_cyl, diameter, stroke, r_pow_cyl, r_diam_strok))
        points.append([diameter, stroke, engines[-1].power, engines[-1].mass, engines[-1].costs])

    costs = list(zip(*points))[-2]
    sorted_idx = (costs.index(cost) for cost in sorted(costs))
    sorted_points = [points[idx] for idx in sorted_idx]
    transposed_points = list(zip(*sorted_points))

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.plot(transposed_points[0], transposed_points[1], transposed_points[-1],
            linestyle='None', marker='o', markersize=0.5)


def three_ways_optimize_engine():
    list_fx = []
    for _ in range(250):
        diameter = opt.BoundedAttributeValue("diameter", 0.05, 0.5)
        stroke = opt.BoundedAttributeValue("stroke", 0.1, 0.3)
        cylinders = opt.FixedAttributeValue("n_cyl", 4)
        r_pow_cyl = opt.FixedAttributeValue("r_pow_cyl", 1e9)
        r_diam_strok = opt.FixedAttributeValue("r_diam_strok", 1.)

        engine_optimizer = EngineOptimizer([cylinders, r_pow_cyl, r_diam_strok], [diameter, stroke])
        model_cma, fx_opt = engine_optimizer.optimize_cma()
        model_grad, fx_opt_grad = engine_optimizer.optimize_gradient()
        model_mix, fx_opt_mix = engine_optimizer.optimize_cma_then_gradient()

        diameters = (x / 1000 for x in range(30, 100, 1))
        strokes = (x / 1000 for x in range(100, 250, 1))
        cylinders = [4]
        list_fx.append(fx_opt_mix)
    return {'cylinders': cylinders, 'diameters': diameters, 'strokes': strokes,
            'model_cma': model_cma, 'model_grad': model_grad, 'model_mix': model_mix,
            'optimal_confs': list_fx}


def test_script():
    results = three_ways_optimize_engine()
    assert(sum(results['optimal_confs']) / len(results['optimal_confs']) <= 0.05)

    assert(opt.Specifications().name == '')
    assert(opt.BoundedAttributeValue('test', 0, 100).dimensionless_value(50) == 0.5)

    assert(opt.Optimizer().adimensioned_vector([1, 2]) is None)
    assert(opt.Optimizer().reduced_vector([1, 2]) is None)
    assert(opt.Optimizer().cma_bounds() is None)
    assert(opt.Optimizer().scipy_minimize_bounds() is None)
    assert(opt.Optimizer().cma_optimization() is None)
    assert(opt.Optimizer().scipy_minimize_optimization() is None)

    assert(opt.DrivenModelOptimizer(lambda x: x**2).model(4) == 16)
    try:
        opt.DrivenModelOptimizer(lambda x: x**2).get_model_from_vector()
    except Exception as e:
        assert(e.args[0] == 'the method must be overloaded by subclassing class')

    standard_optimizer = opt.InstantiatingModelOptimizer([opt.FixedAttributeValue("n_cyl", 4)],
                                                         [opt.BoundedAttributeValue("diameter", 0.05, 0.5)])
    try:
        standard_optimizer.instantiate_model(None)
    except Exception as e:
        assert(e.args[0] == 'the method instantiate_model must be overloaded by subclassing class')
    try:
        standard_optimizer.objective_from_model(lambda x: x**2, clearance=0.003)
    except Exception as e:
        assert(e.args[0] == 'the method objective_from_model must be overloaded by subclassing class')


def __main__():
    results = three_ways_optimize_engine()
    check_costs_function(results['cylinders'], results['diameters'], results['strokes'], 1e8, 1.)

    plt.plot(results['model_cma'].diameter, results['model_cma'].stroke, results['model_cma'].costs,
             linestyle='None', marker='D', markersize=6, color='r', label='Solution CMA')

    plt.plot(results['model_grad'].diameter, results['model_grad'].stroke, results['model_grad'].costs,
             linestyle='None', marker='D', markersize=6, color='m', label='Solution Gradient Descent')

    plt.plot(results['model_mix'].diameter, results['model_mix'].stroke, results['model_mix'].costs,
             linestyle='None', marker='D', markersize=6, color='g', label='Solution approche mixte')

    plt.legend()


# Script
test_script()
if __name__ == "__main__":
    __main__()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Test module for dessia_common. """
from time import sleep
from typing import List, Tuple
import random
import numpy as npy
import matplotlib.pyplot as plt
from dessia_common.core import DessiaObject
import dessia_common.measures as dcm
import dessia_common.files as dcf


class Submodel(DessiaObject):
    """ Mock a MBSE that is a standalone attribute of Model, for testing purpose. """

    _standalone_in_db = True

    def __init__(self, subvalue: int, name: str = ''):
        self.subvalue = subvalue
        self.name = name

        DessiaObject.__init__(self, name=name)


class Model(DessiaObject):
    """ Mock a standalone MBSE for testing purpose. """

    _standalone_in_db = True

    def __init__(self, value: int, submodel: Submodel, name: str = ''):
        self.value = value
        self.submodel = submodel

        DessiaObject.__init__(self, name=name)


class Generator(DessiaObject):
    """ Mock a generator of MBSEs for testing purpose. """

    _standalone_in_db = True
    _allowed_methods = ['long_generation']

    def __init__(self, parameter: int, nb_solutions: int = 25, models: List[Model] = None, name: str = ''):
        self.parameter = parameter
        self.nb_solutions = nb_solutions
        self.models = models

        DessiaObject.__init__(self, name=name)

    def generate(self) -> None:
        """Generate models."""
        submodels = [Submodel(self.parameter * i) for i in range(self.nb_solutions)]
        self.models = [Model(self.parameter + i, submodels[i]) for i in range(self.nb_solutions)]

    def long_generation(self, progress_callback=lambda x: None) -> List[Model]:
        """
        Run a long generation.

        This method aims to test:
            * lots of prints to be caught
            * progress update
            * long computation
        """
        submodels = [Submodel(self.parameter * i) for i in range(self.nb_solutions)]
        models = [Model(self.parameter + i, submodels[i]) for i in range(self.nb_solutions)]
        # Delay to simulate long generation
        print('Beginning a long generation...')
        for i in range(500):
            print(f'Loop n°{i+1} / 500')
            progress = i / 499.
            progress_callback(progress)
            sleep(0.3)
        print('Generation complete')
        return models


class Optimizer(DessiaObject):
    """ Mock an optimizer for testing purpose. """

    _standalone_in_db = True

    def __init__(self, model_to_optimize: Model, name: str = ''):
        self.model_to_optimize = model_to_optimize

        DessiaObject.__init__(self, name=name)

    def optimize(self, optimization_value: int = 3) -> None:
        """Optimize attribute model_to_optimize."""
        self.model_to_optimize.value += optimization_value


class Component(DessiaObject):
    """ MBSE to be tested. Mock a real use case. For example, a module in a battery pack. """

    _standalone_in_db = True

    def __init__(self, efficiency, name: str = ''):
        self.efficiency = efficiency
        DessiaObject.__init__(self, name=name)

    def power_simulation(self, power_value: dcm.Power):
        """ Run a power simulation. """
        return power_value * self.efficiency


class ComponentConnection(DessiaObject):
    """ Mock for a link between to components, e.g. the electrical wiring between two modules in a battery pack. """

    def __init__(self, input_component: Component, output_component: Component, name: str = ''):
        self.input_component = input_component
        self.output_component = output_component
        DessiaObject.__init__(self, name=name)


class SystemUsage(DessiaObject):
    """  Mock a simulation result of a system, e.g. the response of the battery pack to a certain use-case. """

    _standalone_in_db = True

    def __init__(self, time: List[dcm.Time], power: List[dcm.Power], name: str = ''):
        self.time = time
        self.power = power
        DessiaObject.__init__(self, name=name)


class System(DessiaObject):
    """ Mock a system that binds several components and their connection. For example, a battery pack. """

    _standalone_in_db = True
    _dessia_methods = ['power_simulation']

    def __init__(self, components: List[Component], component_connections: List[ComponentConnection], name: str = ''):
        self.components = components
        self.component_connections = component_connections
        DessiaObject.__init__(self, name=name)

    def output_power(self, input_power: dcm.Power):
        """ Compute output power. """
        return input_power * 0.8

    def power_simulation(self, usage: SystemUsage):
        """ Run a power simulation. """
        output_power = []
        for _, input_power in zip(usage.time, usage.power):
            output_power.append(self.output_power(input_power))
        return SystemSimulationResult(self, usage, output_power)


class SystemSimulationResult(DessiaObject):
    """ Result wrapper for a System with its usage. """

    _standalone_in_db = True

    def __init__(self, system: System, system_usage: SystemUsage, output_power: List[dcm.Power], name: str = ''):
        self.system = system
        self.system_usage = system_usage
        self.output_power = output_power
        DessiaObject.__init__(self, name=name)


class SystemSimulationList(DessiaObject):
    """ Results container for system simulation results. """

    _standalone_in_db = True

    def __init__(self, simulations: List[SystemSimulationResult], name: str = ''):
        self.simulations = simulations
        DessiaObject.__init__(self, name=name)


class Car(DessiaObject):
    """ Defines a car with all its features. """

    _standalone_in_db = True
    _non_data_hash_attributes = ['name']

    def __init__(self, name: str, mpg: float, cylinders: int, displacement: dcm.Distance, horsepower: float,
                 weight: dcm.Mass, acceleration: dcm.Time, model: int, origin: str):
        DessiaObject.__init__(self, name=name)

        self.mpg = mpg
        self.cylinders = cylinders
        self.displacement = displacement
        self.horsepower = horsepower
        self.weight = weight
        self.acceleration = acceleration
        self.model = model
        self.origin = origin

    def to_vector(self):
        """ Get equivalent vector of instance of Car. """
        list_formated_car = []
        for feature in self.vector_features():
            list_formated_car.append(getattr(self, feature.lower()))
        return list_formated_car

    @classmethod
    def from_csv(cls, file: dcf.StringFile, end: int = None, remove_duplicates: bool = False):
        """ Generates Cars from given .csv file. """
        array = [row.split(',') for row in file.getvalue().split('\n')][1:-1]
        cars = []
        for idx_line, line in enumerate(array):
            if end is not None and idx_line >= end:
                break
            if not remove_duplicates or (remove_duplicates and line.tolist() not in cars):
                attr_list = [float(attr) if attr.replace('.', '').isnumeric() else attr for attr in line]
                attr_list[3] /= 1000

                for idx_attr, attr in enumerate(attr_list):
                    if isinstance(attr, npy.int64):
                        attr_list[idx_attr] = int(attr)
                    elif isinstance(attr, npy.float64):
                        attr_list[idx_attr] = float(attr)
                cars.append(cls(*attr_list))
        return cars


class CarWithFeatures(Car):
    """ Defines a car with all specific features (`'mpg', 'displacement', 'horsepower', 'acceleration', 'weight'`). """

    _vector_features = ['mpg', 'displacement', 'horsepower', 'acceleration', 'weight']

    def __init__(self, name: str, mpg: float, cylinders: int, displacement: dcm.Distance, horsepower: float,
                 weight: dcm.Mass, acceleration: dcm.Time, model: int, origin: str):
        Car.__init__(self, name, mpg, cylinders, displacement, horsepower, weight, acceleration, model, origin)

    @classmethod
    def vector_features(cls):
        """ Get list of _vector_features. """
        return cls._vector_features


class RandDataD1(DessiaObject):
    """ Creates a dataset with 1 parameters from a number of clusters and dimensions. """

    _standalone_in_db = True
    _non_data_hash_attributes = ['name']
    _nb_dims = 1
    _vector_features = [f'p_{i+1}' for i in range(_nb_dims)]

    def __init__(self, p_1: float, name: str = ''):
        DessiaObject.__init__(self, name=name)
        self.p_1 = p_1

    @classmethod
    def vector_features(cls):
        """Get list of _vector_features."""
        return cls._vector_features

    @classmethod
    def create_dataset(cls, nb_clusters: int = 10, nb_points: int = 2500, mean_borns: Tuple[float, float] = (-50., 50),
                       std_borns: Tuple[float, float] = (-2., 2.)):
        """ Create a random dataset with a number of clusters, number of points, means and std per cluster. """
        means_list = []
        std_list = []
        data_list = []
        cluster_sizes = cls.set_cluster_sizes(nb_points, nb_clusters)

        for cluster_size in cluster_sizes:
            means_list = [random.uniform(*mean_borns) for i in range(cls._nb_dims)]
            std_list = [random.uniform(*std_borns) for i in range(cls._nb_dims)]
            for _ in range(cluster_size):
                data_list.append(
                    cls(*[random.normalvariate(means_list[dim], std_list[dim]) for dim in range(cls._nb_dims)]))

        return data_list

    @staticmethod
    def set_cluster_sizes(nb_points: int, nb_clusters: int):
        """ Set clusters' sizes in function of total number of points. """
        current_nb_points = nb_points
        cluster_sizes = []
        for _ in range(nb_clusters - 1):
            points_in_cluster = random.randint(int(current_nb_points / nb_clusters / 2),
                                               int(current_nb_points / nb_clusters * 2))
            cluster_sizes.append(points_in_cluster)
            current_nb_points -= points_in_cluster

        cluster_sizes.append(int(nb_points - npy.sum(cluster_sizes)))
        return cluster_sizes

    @staticmethod
    def python_plot(x_label: str, y_label: str, rand_data_list: List[List[float]], **kwargs):
        """Plot with matplotlib."""
        x_coords = [getattr(RandData, x_label) for RandData in rand_data_list]
        y_coords = [getattr(RandData, y_label) for RandData in rand_data_list]
        plt.plot(x_coords, y_coords, **kwargs)


class RandDataD2(RandDataD1):
    """ Creates a dataset with 2 parameters from a number of clusters and dimensions. """

    _nb_dims = 2
    _vector_features = [f'p_{i+1}' for i in range(2)]

    def __init__(self, p_1: float, p_2: float, name: str = ''):
        RandDataD1.__init__(self, p_1, name=name)
        self.p_2 = p_2


class RandDataD3(RandDataD1):
    """ Creates a dataset with 3 parameters from a number of clusters and dimensions. """

    _nb_dims = 3
    _vector_features = [f'p_{i+1}' for i in range(3)]

    def __init__(self, p_1: float, p_2: float, p_3: float, name: str = ''):
        RandDataD1.__init__(self, p_1, name=name)
        self.p_2 = p_2
        self.p_3 = p_3


class RandDataD4(RandDataD1):
    """ Creates a dataset with 4 parameters from a number of clusters and dimensions. """

    _nb_dims = 4
    _vector_features = [f'p_{i+1}' for i in range(4)] + ['test_prop']

    def __init__(self, p_1: float, p_2: float, p_3: float, p_4: float, name: str = ''):
        RandDataD1.__init__(self, p_1, name=name)
        self.p_2 = p_2
        self.p_3 = p_3
        self.p_4 = p_4
        self._test_prop = None

    @property
    def test_prop(self):
        """ Fake property for some tests. """
        if self._test_prop is None:
            self._test_prop = 3
        return self._test_prop


class RandDataD5(RandDataD1):
    """ Creates a dataset with 5 parameters from a number of clusters and dimensions. """

    _nb_dims = 5
    _vector_features = [f'p_{i+1}' for i in range(5)] + ['test_prop']

    def __init__(self, p_1: float, p_2: float, p_3: float, p_4: float, p_5: float, name: str = ''):
        RandDataD1.__init__(self, p_1, name=name)
        self.p_2 = p_2
        self.p_3 = p_3
        self.p_4 = p_4
        self.p_5 = p_5
        self._test_prop = None

    @property
    def test_prop(self):
        """Dummy property for some tests."""
        if self._test_prop is None:
            self._test_prop = 3
        return self._test_prop


class RandDataD6(RandDataD1):
    """ Creates a dataset with 6 parameters from a number of clusters and dimensions. """

    _nb_dims = 6
    _vector_features = [f'p_{i+1}' for i in range(6)] + ['test_prop']

    def __init__(self, p_1: float, p_2: float, p_3: float, p_4: float, p_5: float, p_6: float, name: str = ''):
        RandDataD1.__init__(self, p_1, name=name)
        self.p_2 = p_2
        self.p_3 = p_3
        self.p_4 = p_4
        self.p_5 = p_5
        self.p_6 = p_6
        self._test_prop = None

    @property
    def test_prop(self):
        """ Dummy property for some tests. """
        if self._test_prop is None:
            self._test_prop = 3
        return self._test_prop


class RandDataD7(RandDataD1):
    """ Creates a dataset with 7 parameters from a number of clusters and dimensions. """

    _nb_dims = 7
    _vector_features = [f'p_{i+1}' for i in range(7)]
    def __init__(self, p_1: float, p_2: float, p_3: float, p_4: float, p_5: float, p_6: float, p_7: float,
                 name: str = ''):
        RandDataD1.__init__(self, p_1, name=name)
        self.p_2 = p_2
        self.p_3 = p_3
        self.p_4 = p_4
        self.p_5 = p_5
        self.p_6 = p_6
        self.p_7 = p_7


class RandDataD8(RandDataD1):
    """ Creates a dataset with 8 parameters from a number of clusters and dimensions. """

    _nb_dims = 8
    _vector_features = [f'p_{i+1}' for i in range(8)]
    def __init__(self, p_1: float, p_2: float, p_3: float, p_4: float, p_5: float, p_6: float, p_7: float, p_8: float,
                 name: str = ''):
        RandDataD1.__init__(self, p_1, name=name)
        self.p_2 = p_2
        self.p_3 = p_3
        self.p_4 = p_4
        self.p_5 = p_5
        self.p_6 = p_6
        self.p_7 = p_7
        self.p_8 = p_8


class RandDataD9(RandDataD1):
    """ Creates a dataset with 9 parameters from a number of clusters and dimensions. """

    _nb_dims = 9
    _vector_features = [f'p_{i+1}' for i in range(9)]
    def __init__(self, p_1: float, p_2: float, p_3: float, p_4: float, p_5: float, p_6: float, p_7: float, p_8: float,
                 p_9: float, name: str = ''):
        RandDataD1.__init__(self, p_1, name=name)
        self.p_2 = p_2
        self.p_3 = p_3
        self.p_4 = p_4
        self.p_5 = p_5
        self.p_6 = p_6
        self.p_7 = p_7
        self.p_8 = p_8
        self.p_9 = p_9


class RandDataD10(RandDataD1):
    """ Creates a dataset with 10 parameters from a number of clusters and dimensions. """

    _nb_dims = 10
    _vector_features = [f'p_{i+1}' for i in range(10)]

    def __init__(self, p_1: float, p_2: float, p_3: float, p_4: float, p_5: float, p_6: float, p_7: float, p_8: float,
                 p_9: float, p_10: float, name: str = ''):
        RandDataD1.__init__(self, p_1, name=name)
        self.p_2 = p_2
        self.p_3 = p_3
        self.p_4 = p_4
        self.p_5 = p_5
        self.p_6 = p_6
        self.p_7 = p_7
        self.p_8 = p_8
        self.p_9 = p_9
        self.p_10 = p_10



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Write document file. """

import re
from typing import List, Union
import docx

from dessia_common.files import BinaryFile


class LayoutElement:
    """
    Represents a header or footer in document.

    This class is used to create headers and footers for document. Subclasses
    `Header` and `Footer` are available for convenience.
    """

    def __init__(self, text: str, align: str = 'center'):
        self.text = text
        self.align = align

    def _add_to_section(self, section, type_: str):
        """
        Add the header or footer to the specified section.

        :param section: The `docx.section.Section` object to add the header/footer to.
        :param type_: The type of element to add, either "header" or "footer".
        """
        element = getattr(section, type_)
        paragraph = element.add_paragraph()
        paragraph.text = self.text
        paragraph.alignment = getattr(docx.enum.text.WD_PARAGRAPH_ALIGNMENT, self.align.upper())

    def _add_picture(self, section, type_: str, image_path: str, width: int = None, height: int = None):
        """ Add the header or footer picture to the specified section. """
        element = getattr(section, type_)
        paragraph = element.paragraphs[0] if element.paragraphs else element.add_paragraph()
        run = paragraph.add_run()
        run.add_picture(image_path, width=width, height=height)


class Header(LayoutElement):
    """ Represents a header in document. """

    def add_to_document(self, document: docx.Document):
        """ Add the header to the document. """
        section = document.sections[0]
        super()._add_to_section(section=section, type_='header')

    def add_to_section(self, section):
        """ Add header to section. """
        super()._add_to_section(section=section, type_='header')

    def add_picture(self, section, image_path: str, width: int = None, height: int = None):
        """ Add picture for header. """
        super()._add_picture(section=section, type_='header', image_path=image_path, width=width, height=height)


class Footer(LayoutElement):
    """ Represents a footer in document. """

    def add_to_document(self, document: docx.Document):
        """ Add the footer to the document. """
        section = document.sections[0]
        super()._add_to_section(section=section, type_='footer')

    def add_to_section(self, section):
        """ Add footer to section. """
        super()._add_to_section(section=section, type_='footer')

    def add_picture(self, section, image_path: str, width: int = None, height: int = None):
        """ Add picture for footer. """
        super()._add_picture(section=section, type_='footer', image_path=image_path, width=width, height=height)


class Heading:
    """ Represents a heading in document. """

    def __init__(self, text: str, level: int):
        self.text = text
        self.level = level

    def add_to_document(self, document: docx.Document):
        """ Add the heading to the document. """
        document.add_heading(self.text, self.level)

    @classmethod
    def from_markdown(cls, markdown_text: str):
        """ Create a new Paragraph object from Markdown text. """
        return cls(text=markdown_text.strip("#"), level=markdown_text.count("#"))


class Paragraph:
    """ Represents a paragraph in the document. """

    def __init__(self, text: str):
        self.text = text

    def add_to_document(self, document: docx.Document):
        """ Add paragraph to the document. """
        document.add_paragraph(self.text)

    @classmethod
    def from_markdown(cls, markdown_text: str):
        """ Create a new Paragraph object from Markdown text. """
        return cls(text=markdown_text)


class Table:
    """ Represents a table in the document. """

    def __init__(self, rows: List[List[str]]):
        self.rows = rows

    def add_to_document(self, document: docx.Document, style: str = "TableGrid"):
        """
        Add the table to the document.

        :param document: The document to which the table will be added.
        :param style: The table style to use. Default is "TableGrid". Other available styles include:
                        - "LightShading": adds light shading to the table cells.
                        - "LightGrid": similar to "TableGrid", but with lighter grid lines.
                        - "MediumShading1", "MediumShading2": adds medium-level shading to the table cells,
                        with different levels of intensity.
                        - "DarkList": displays the table as a numbered list, with a dark background color for each item.
                        - "ColorfulGrid": uses bright, contrasting colors for the table borders and cell backgrounds.
                        - "LightList": displays the table as a bulleted list, with a light background color for each
                        item.
        """
        table = document.add_table(rows=len(self.rows), cols=len(self.rows[0]), style=style)
        for i, row in enumerate(self.rows):
            for j, cell in enumerate(row):
                table.cell(i, j).text = cell


class Section:
    """
    Represents a section in the document.

    A section is a part of the document with its own headers, footers.
    """

    def __init__(self):
        self.elements = []

    def add_element(self, element: Union[Header, Footer]):
        """ Add a header to the section. """
        self.elements.append(element)

    def add_to_document(self, document: docx.Document):
        """ Add the section to the document. """
        section = document.sections[0]
        for element in self.elements:
            element.add_to_section(section)

    def add_picture_to_document(self, document: docx.Document, image_path, width, height):
        """ Add picture to the section. """
        section = document.sections[0]
        if self.elements:
            self.elements[0].add_picture(section, image_path, width, height)


class DocxWriter:
    """ Write a Document file. """

    def __init__(self, paragraphs: List[Paragraph] = None, section: Section = None, filename: str = "document.docx",
                 headings: List[Heading] = None, tables: List[Table] = None):
        self.filename = filename
        if headings is None:
            headings = []
        self.headings = headings
        self.section = section
        if paragraphs is None:
            paragraphs = []
        self.paragraphs = paragraphs
        self.tables = tables
        self.document = docx.Document()

        if self.section:
            self.section.add_to_document(document=self.document)

    def add_headings(self, page_break: bool = True) -> 'DocxWriter':
        """ Add a list of headings to the document. """
        for heading in self.headings:
            self.document.add_heading(heading.text, level=heading.level)
        if page_break:
            self.add_page_breaks(num_page_breaks=1)
        return self

    def add_paragraphs(self, add_heading: bool = True) -> 'DocxWriter':
        """ Add a list of paragraphs to the document. """
        for paragraph in self.paragraphs:
            if self.headings and add_heading:
                self.document.add_heading(self.headings[0].text, self.headings[0].level)
                self.headings = self.headings[1:]
            self.document.add_paragraph(paragraph.text)
        return self

    def add_paragraph_as_heading(self, text: str):
        """
        Adds a new heading to the document, using the specified text as the heading text.

        The level of the heading is determined by the number of '#' characters in the text.
        For example, a single '#' character at the beginning of the text will create a level 1 heading.
        """
        self.document.add_heading(text, level=text.count('#'))
        return self

    def add_page_breaks(self, num_page_breaks: int):
        """ Add an empty page to the document. """
        for _ in range(num_page_breaks):
            self.document.add_page_break()

    def add_table(self, add_all_tables: bool = False, table_index: int = 0) -> 'DocxWriter':
        """ Add tables to the document. """
        if add_all_tables:
            for table in self.tables:
                table.add_to_document(self.document)
                self.document.add_paragraph()
        else:
            self.tables[table_index].add_to_document(self.document)
        return self

    def add_list_items(self, items: List[str], style: str = 'List Bullet'):
        """
        Add a list (bullet or numbered) to the document.

        :param items: A list of strings to be added as list items
        :type items: List[str]

        :param style: The style of the list. Valid values are 'List Bullet' (default) or 'List Number'
        :type style: str
        """
        for item in items:
            self.document.add_paragraph(item, style=style)
        return self

    def add_header_footer_picture(self, image_path: str, width: int = None, height: int = None)\
            -> 'DocxWriter':
        """ Add a picture to the header or footer of the document. """
        self.section.add_picture_to_document(document=self.document, image_path=image_path, width=width, height=height)
        return self

    def add_picture(self, image_path: str, width: int = None, height: int = None) -> 'DocxWriter':
        """ Add an image to the document. """
        self.document.add_picture(image_path, width=width, height=height)
        return self

    def delete_layout_element(self):
        """ Remove the header and footer from all sections in a Word document. """
        for section in self.document.sections:
            section.different_first_page_header_footer = False
            section.header.is_linked_to_previous = True
            section.footer.is_linked_to_previous = True

    @staticmethod
    def parse_markdown(markdown_text: str):
        """
        Parses the given markdown text.

        :return: Tuple containing a list of headings, a list of paragraphs and tables.
        """
        elements, headings = [], []
        table_pattern = re.compile(r'^\|.*\|$')
        horizontal_line_pattern = re.compile(r'^\s*\|?\s*-+\s*\|?\s*(-+\s*\|?)*\s*$')

        for line in markdown_text.split('\n'):
            line = line.strip()
            if line:
                if line.startswith('#'):
                    headings.append(Heading.from_markdown(line))
                    elements.append(Paragraph.from_markdown(markdown_text=line))

                elif table_pattern.match(line) and not horizontal_line_pattern.match(line):
                    row = line.strip('|').split('|')
                    if '---' not in line:
                        elements.append(Table(rows=[row]))

                else:
                    if '---' not in line:
                        elements.append(Paragraph.from_markdown(markdown_text=line))
        return headings, elements

    @classmethod
    def from_markdown(cls, markdown_text: str):
        """ Converts the given markdown text into a DocxWriter instance."""
        headings, elements = cls.parse_markdown(markdown_text=markdown_text)
        docx_writer = cls(headings=headings)
        docx_writer.add_headings()
        for item in elements:
            if isinstance(item, Paragraph):
                docx_writer.paragraphs = [item]
                if item.text.startswith('#'):
                    docx_writer.add_paragraph_as_heading(text=item.text.strip(' # '))
                else:
                    docx_writer.add_paragraphs(add_heading=False)
            if isinstance(item, Table):
                docx_writer.tables = [item]
                docx_writer.add_table()
        return docx_writer

    def save_file(self):
        """ Saves the document to a file. """
        if not self.filename.endswith('.docx'):
            self.filename += '.docx'
        self.document.save(self.filename)

    def save_to_stream(self, stream: BinaryFile):
        """ Saves the document to a binary stream. """
        self.document.save(stream)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""

import json

from dessia_common import REF_MARKER
from dessia_common.models.workflows import simulation_workflow
from dessia_common.models.power_test import components1, component_connections1, usage1

simulation_workflow.to_dict(use_pointers=False)
input_values = {0: components1, 1: component_connections1, 3: usage1}

workflow_run = simulation_workflow.run(input_values)


# Testing that there is no pointer when use_pointers=False
d = workflow_run.to_dict(use_pointers=False)
s = json.dumps(d)
if REF_MARKER in s:
    ind_ref = s.index(REF_MARKER)
    print(s[ind_ref - 300:ind_ref + 500])
    raise ValueError('Pointer detected with use_pointers=False')

simulation_workflow.save_script_to_file('_simulation_workflow')

print("script power_simulation.py has passed")



"""
Test for ClusteredDataset in workflow. Test filtering method
"""
import json
import importlib_resources
from dessia_common.files import StringFile
from dessia_common.typings import ClassMethodType, MethodType
from dessia_common.tests import Car
from dessia_common.workflow.blocks import ClassMethod, InstantiateModel, ModelMethod, Unpacker, Concatenate
from dessia_common.core import DessiaFilter, FiltersList
from dessia_common.datatools.dataset import Dataset
from dessia_common.datatools.cluster import ClusteredDataset
from dessia_common.workflow.core import Workflow, Pipe

# Import data
ref = importlib_resources.files("dessia_common").joinpath("models/data/cars.csv")
with ref.open('rb') as csv_cars:
    stream = StringFile.from_stream(csv_cars)

    # ===============================================================================================================
    # ClusteredDataset Workflow
    # ===============================================================================================================

    block_0 = ClassMethod(method_type=ClassMethodType(Car, 'from_csv'), name='CSV Cars')
    block_1 = InstantiateModel(model_class=Dataset, name='HList Cars')
    block_2 = ClassMethod(method_type=ClassMethodType(FiltersList, 'from_filters_list'), name='Filters Clist')
    block_3 = ClassMethod(method_type=ClassMethodType(ClusteredDataset, 'from_agglomerative_clustering'), name='Clustering')
    block_4 = ModelMethod(method_type=MethodType(ClusteredDataset, 'filtering'), name='CList.filtering')
    blocks = [block_0, block_1, block_2, block_3, block_4]

    pipe_0 = Pipe(block_0.outputs[0], block_1.inputs[0])
    pipe_1 = Pipe(block_1.outputs[0], block_3.inputs[0])
    pipe_2 = Pipe(block_3.outputs[0], block_4.inputs[0])
    pipe_3 = Pipe(block_2.outputs[0], block_4.inputs[1])
    pipes = [pipe_0, pipe_1, pipe_2, pipe_3]

    workflow = Workflow(blocks, pipes, output=block_4.outputs[0], name='Filters demo Clists')

    # Workflow run
    filters = [DessiaFilter('cylinders', ">", 6), DessiaFilter('displacement', ">", 0.3)]
    workflow_run = workflow.run({workflow.input_index(block_0.inputs[0]): stream,
                                 workflow.input_index(block_2.inputs[0]): filters,
                                 workflow.input_index(block_2.inputs[1]): 'and',
                                 workflow.input_index(block_3.inputs[1]): 10})

    # Workflow tests
    workflow._check_platform()
    wfrun_plot_data = workflow_run.output_value.plot_data()

    # JSON TESTS
    output_dict = workflow_run.output_value[[0, 3, 10, 15, 30, -1]].to_dict(use_pointers=True)
    output_json = json.dumps(output_dict)
    output_json_to_dict = json.loads(output_json)
    output_jsondict_to_object_1 = ClusteredDataset.dict_to_object(output_json_to_dict)

    # ===============================================================================================================
    # ClusteredDataset Big Workflow
    # ===============================================================================================================

    block_0 = ClassMethod(method_type=ClassMethodType(Car, 'from_csv'), name='CSV Cars')
    block_1 = InstantiateModel(model_class=Dataset, name='HList Cars')
    block_2 = ClassMethod(method_type=ClassMethodType(ClusteredDataset, 'from_agglomerative_clustering'), name='Clustering')
    block_3 = ModelMethod(method_type=MethodType(ClusteredDataset, 'clustered_sublists'), name='Sublists')
    block_4 = Unpacker(indices=[3, 8, 9], name='Unpack_3_8_9')
    block_5 = Concatenate(number_arguments=3, name='Concatenate')
    block_6 = ClassMethod(method_type=ClassMethodType(ClusteredDataset, 'from_dbscan'), name='DBSCAN')
    blocks = [block_0, block_1, block_2, block_3, block_4, block_5, block_6]

    pipe_0 = Pipe(block_0.outputs[0], block_1.inputs[0])
    pipe_1 = Pipe(block_1.outputs[0], block_2.inputs[0])
    pipe_2 = Pipe(block_2.outputs[0], block_3.inputs[0])
    pipe_3 = Pipe(block_3.outputs[0], block_4.inputs[0])
    pipe_4 = Pipe(block_4.outputs[0], block_5.inputs[0])
    pipe_5 = Pipe(block_4.outputs[1], block_5.inputs[1])
    pipe_6 = Pipe(block_4.outputs[2], block_5.inputs[2])
    pipe_7 = Pipe(block_5.outputs[0], block_6.inputs[0])
    pipes = [pipe_0, pipe_1, pipe_2, pipe_3, pipe_4, pipe_5, pipe_6, pipe_7]

    workflow = Workflow(blocks, pipes, output=block_6.outputs[0], name='SubClist + Concatenate')

    # Workflow run
    workflow_run = workflow.run({workflow.input_index(block_0.inputs[0]): stream,
                                 workflow.input_index(block_2.inputs[1]): 10,
                                 workflow.input_index(block_6.inputs[1]): 40})

# Workflow tests
workflow._check_platform()
wfrun_plot_data = workflow_run.output_value.plot_data()

# JSON TESTS
output_dict = workflow_run.output_value[[0, 3, 5, 2, 12, -1]].to_dict(use_pointers=True)
output_json = json.dumps(output_dict)
output_json_to_dict = json.loads(output_json)
output_jsondict_to_object_2 = ClusteredDataset.dict_to_object(output_json_to_dict)
print(output_jsondict_to_object_2)

# ===============================================================================================================
# TESTS ON RESULTS
# ===============================================================================================================

reference_output_1 = {
  "name": "",
  "object_class": "dessia_common.datatools.cluster.ClusteredDataset",
  "package_version": "0.9.2.dev320+gecd4609",
  "dessia_objects": [
    {
      "name": "Chevrolet Chevelle Malibu",
      "object_class": "dessia_common.tests.Car",
      "package_version": "0.9.2.dev320+gecd4609",
      "mpg": 18.0,
      "cylinders": 8.0,
      "displacement": 0.307,
      "horsepower": 130.0,
      "weight": 3504.0,
      "acceleration": 12.0,
      "model": 70.0,
      "origin": "US\r"
    },
    {
      "name": "AMC Rebel SST",
      "object_class": "dessia_common.tests.Car",
      "package_version": "0.9.2.dev320+gecd4609",
      "mpg": 16.0,
      "cylinders": 8.0,
      "displacement": 0.304,
      "horsepower": 150.0,
      "weight": 3433.0,
      "acceleration": 12.0,
      "model": 70.0,
      "origin": "US\r"
    },
    {
      "name": "Chevrolet Chevelle Concours (sw)",
      "object_class": "dessia_common.tests.Car",
      "package_version": "0.9.2.dev320+gecd4609",
      "mpg": 0.0,
      "cylinders": 8.0,
      "displacement": 0.35,
      "horsepower": 165.0,
      "weight": 4142.0,
      "acceleration": 11.5,
      "model": 70.0,
      "origin": "US\r"
    },
    {
      "name": "Plymouth 'Cuda 340",
      "object_class": "dessia_common.tests.Car",
      "package_version": "0.9.2.dev320+gecd4609",
      "mpg": 14.0,
      "cylinders": 8.0,
      "displacement": 0.34,
      "horsepower": 160.0,
      "weight": 3609.0,
      "acceleration": 8.0,
      "model": 70.0,
      "origin": "US\r"
    },
    {
      "name": "Chevrolet Impala",
      "object_class": "dessia_common.tests.Car",
      "package_version": "0.9.2.dev320+gecd4609",
      "mpg": 13.0,
      "cylinders": 8.0,
      "displacement": 0.35,
      "horsepower": 165.0,
      "weight": 4274.0,
      "acceleration": 12.0,
      "model": 72.0,
      "origin": "US\r"
    },
    {
      "name": "Oldsmobile Cutlass LS",
      "object_class": "dessia_common.tests.Car",
      "package_version": "0.9.2.dev320+gecd4609",
      "mpg": 26.6,
      "cylinders": 8.0,
      "displacement": 0.35,
      "horsepower": 105.0,
      "weight": 3725.0,
      "acceleration": 19.0,
      "model": 81.0,
      "origin": "US\r"
    }],
  "labels": [1, 0, 4, 1, 4, 1]}

reference_output_2 = {
  'name': '',
  'object_class': 'dessia_common.datatools.cluster.ClusteredDataset',
  'package_version': '0.8.1.dev105+gb6cb82bfeatcluster',
  'dessia_objects': [
    {
      'name': 'Volkswagen 1131 Deluxe Sedan',
      'object_class': 'dessia_common.tests.Car',
      'package_version': '0.8.1.dev105+gb6cb82bfeatcluster',
      'mpg': 26.0,
      'cylinders': 4.0,
      'displacement': 0.097,
      'horsepower': 46.0,
      'weight': 1835.0,
      'acceleration': 20.5,
      'model': 70.0,
      'origin': 'Europe\r'
    },
    {
     'name': 'Volkswagen Super Beetle 117',
     'object_class': 'dessia_common.tests.Car',
     'package_version': '0.8.1.dev105+gb6cb82bfeatcluster',
     'mpg': 0.0,
     'cylinders': 4.0,
     'displacement': 0.097,
     'horsepower': 48.0,
     'weight': 1978.0,
     'acceleration': 20.0,
     'model': 71.0,
     'origin': 'Europe\r'},
  {
     'name': 'Peugeot 304',
     'object_class': 'dessia_common.tests.Car',
     'package_version': '0.8.1.dev105+gb6cb82bfeatcluster',
     'mpg': 30.0,
     'cylinders': 4.0,
     'displacement': 0.079,
     'horsepower': 70.0,
     'weight': 2074.0,
     'acceleration': 19.5,
     'model': 71.0,
     'origin': 'Europe\r'},
  {
     'name': 'Toyota Corolla 1200',
     'object_class': 'dessia_common.tests.Car',
     'package_version': '0.8.1.dev105+gb6cb82bfeatcluster',
     'mpg': 31.0,
     'cylinders': 4.0,
     'displacement': 0.071,
     'horsepower': 65.0,
     'weight': 1773.0,
     'acceleration': 19.0,
     'model': 71.0,
     'origin': 'Japan\r'},
  {
     'name': 'Toyota Corolla 1200',
     'object_class': 'dessia_common.tests.Car',
     'package_version': '0.8.1.dev105+gb6cb82bfeatcluster',
     'mpg': 32.0,
     'cylinders': 4.0,
     'displacement': 0.071,
     'horsepower': 65.0,
     'weight': 1836.0,
     'acceleration': 21.0,
     'model': 74.0,
     'origin': 'Japan\r'},
  {
     'name': 'Chevy S-10',
     'object_class': 'dessia_common.tests.Car',
     'package_version': '0.8.1.dev105+gb6cb82bfeatcluster',
     'mpg': 31.0,
     'cylinders': 4.0,
     'displacement': 0.119,
     'horsepower': 82.0,
     'weight': 2720.0,
     'acceleration': 19.4,
     'model': 82.0,
     'origin': 'US\r'
     }],
  'labels': [0, 1, 1, 0, 0, 4]}

# Tests
assert(output_jsondict_to_object_1 == ClusteredDataset.dict_to_object(reference_output_1))
assert(output_jsondict_to_object_2 == ClusteredDataset.dict_to_object(reference_output_2))

